"""
Pillar data from vCenter or an ESXi host

.. versionadded:: 2017.7.0

:depends: - pyVmomi

This external pillar can pull attributes from objects in vCenter or an ESXi host and provide those attributes
as pillar data to minions.  This can allow for pillar based targeting of minions on ESXi host, Datastore, VM
configuration, etc.  This setup requires only the salt master have access to the vCenter server/ESXi hosts.

The pillar will return an empty dict if the 'os' or 'virtual' grain are not 'VMWare', 'ESXi', or 'VMWare ESXi'.

Defaults
========

- The external pillar will search for Virtual Machines with the VM name matching the minion id.
- Data will be returned into the 'vmware' pillar key.
- The external pillar has a default set of properties to return for both VirtualMachine and HostSystem types.


Configuring the VMWare pillar
=============================

The required minimal configuration in the salt master ext_pillar setup:

.. code-block:: yaml

    ext_pillar:
        - vmware:
            host: <vcenter/esx host>
            username: <user to connect with>
            password: <password>

Optionally, the following keyword arguments can be passed to the ext_pillar for customized configuration:

    pillar_key
        Optionally set the pillar key to return the data into.  Default is ``vmware``.

    protocol
        Optionally set to alternate protocol if the vCenter server or ESX/ESXi host is not
        using the default protocol. Default protocol is ``https``.

    port
        Optionally set to alternate port if the vCenter server or ESX/ESXi host is not
        using the default port. Default port is ``443``.

    property_name
        Property name to match the minion id against.  Defaults to ``name``.

    property_types
        Optionally specify a list of pyVmomi vim types to search for the minion id in 'property_name'.
        Default is ``['VirtualMachine']``.

        For example, to search both vim.VirtualMachine and vim.HostSystem object types:

        .. code-block:: yaml

           ext_pillar:
               - vmware:
                   host: myesx
                   username: root
                   password: complex_password
                   property_types:
                     - VirtualMachine
                     - HostSystem

        Additionally, the list of property types can be dicts, the item of the dict being a list specifying
        the attribute to return for that vim object type.

        The pillar will attempt to recurse the attribute and return all child attributes.

        To explicitly specify deeper attributes without attempting to recurse an attribute, convert the list
        item to a dict with the item of the dict being the child attributes to return.  Follow this pattern
        to return attributes as deep within the object as necessary.

        .. note::
            Be careful when specifying custom attributes!  Many attributes have objects as attributes which
            have the parent object as an attribute and which will cause the pillar to fail due to the attempt
            to convert all sub-objects recursively (i.e. infinite attribute loops).  Specifying only the
            sub-attributes you would like returned will keep the infinite recursion from occurring.

            A maximum recursion exception will occur in this case and the pillar will not return as desired.

        .. code-block:: yaml

            ext_pillar:
                - vmware:
                    host: myvcenter
                    username: my_user
                    password: my_pass
                    replace_default_attributes: True
                    property_types:
                      - VirtualMachine:
                          - config:
                             - bootOptions:
                                 - bootDelay
                                 - bootRetryDelay
                      - HostSystem:
                          - datastore:
                             - name

        The above ext_pillar example would return a pillar like the following for a VirtualMachine object that's
        name matched the minion id:

        .. code-block:: yaml

            vmware:
              config:
                bootOptions:
                  bootDelay: 1000
                  bootRetryDelay: 1000

        If you were to retrieve these virtual machine attributes via pyVmomi directly, this would be the same as

        .. code-block:: python

            vmObject.config.bootOptions.bootDelay
            vmObject.config.bootOptionis.bootRetryDelay

        The above ext_pillar example would return a pillar like the following for a HostySystem object that's name
        matched the minion id:

        .. code-block:: yaml

            vmware:
              datastore:
                 - name: Datastore1
                 - name: Datastore2

        The 'datastore' property of a HostSystem object is a list of datastores, thus a list is returned.

    replace_default_attributes
        If custom attributes are specified by the property_types parameter, replace_default_attributes determines
        if those will be added to default attributes (False) or replace the default attributes completely (True).
        The default setting is 'False'.

        .. note::

            vCenter "Custom Attributes" (i.e. Annotations) will always be returned if it exists on the object as
            part of the pillar regardless of this setting.

"""

import logging

import salt.utils.dictupdate as dictupdate
import salt.utils.vmware

try:
    # pylint: disable=no-name-in-module
    from pyVim.connect import Disconnect
    from pyVmomi import vim

    HAS_LIBS = True
    # pylint: enable=no-name-in-module
except ImportError:
    HAS_LIBS = False


__virtualname__ = "vmware"

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only return if python-etcd is installed
    """
    return __virtualname__ if HAS_LIBS else False


def ext_pillar(minion_id, pillar, **kwargs):  # pylint: disable=W0613
    """
    Check vmware/vcenter for all data
    """
    vmware_pillar = {}
    host = None
    username = None
    password = None
    property_types = []
    property_name = "name"
    protocol = None
    port = None
    pillar_key = "vmware"
    replace_default_attributes = False
    type_specific_pillar_attributes = {
        "VirtualMachine": [
            {
                "config": [
                    "version",
                    "guestId",
                    "files",
                    "tools",
                    "flags",
                    "memoryHotAddEnabled",
                    "cpuHotAddEnabled",
                    "cpuHotRemoveEnabled",
                    "datastoreUrl",
                    "swapPlacement",
                    "bootOptions",
                    "scheduledHardwareUpgradeInfo",
                    "memoryAllocation",
                    "cpuAllocation",
                ]
            },
            {
                "summary": [
                    {"runtime": [{"host": ["name", {"parent": "name"}]}, "bootTime"]},
                    {
                        "guest": [
                            "toolsStatus",
                            "toolsVersionStatus",
                            "toolsVersionStatus2",
                            "toolsRunningStatus",
                        ]
                    },
                    {"config": ["cpuReservation", "memoryReservation"]},
                    {"storage": ["committed", "uncommitted", "unshared"]},
                    {"dasVmProtection": ["dasProtected"]},
                ]
            },
            {
                "storage": [
                    {
                        "perDatastoreUsage": [
                            {"datastore": "name"},
                            "committed",
                            "uncommitted",
                            "unshared",
                        ]
                    }
                ]
            },
        ],
        "HostSystem": [
            {
                "datastore": [
                    "name",
                    "overallStatus",
                    {
                        "summary": [
                            "url",
                            "freeSpace",
                            "maxFileSize",
                            "maxVirtualDiskCapacity",
                            "maxPhysicalRDMFileSize",
                            "maxVirtualRDMFileSize",
                            {
                                "vmfs": [
                                    "capacity",
                                    "blockSizeMb",
                                    "maxBlocks",
                                    "majorVersion",
                                    "version",
                                    "uuid",
                                    {"extent": ["diskName", "partition"]},
                                    "vmfsUpgradeable",
                                    "ssd",
                                    "local",
                                ],
                            },
                        ],
                    },
                    {"vm": "name"},
                ]
            },
            {
                "vm": [
                    "name",
                    "overallStatus",
                    {"summary": [{"runtime": "powerState"}]},
                ]
            },
        ],
    }
    pillar_attributes = [
        {"summary": ["overallStatus"]},
        {"network": ["name", {"config": {"distributedVirtualSwitch": "name"}}]},
        {"datastore": ["name"]},
        {"parent": ["name"]},
    ]

    if "pillar_key" in kwargs:
        pillar_key = kwargs["pillar_key"]
    vmware_pillar[pillar_key] = {}

    if "host" not in kwargs:
        log.error(
            "VMWare external pillar configured but host is not specified in ext_pillar"
            " configuration."
        )
        return vmware_pillar
    else:
        host = kwargs["host"]
        log.debug("vmware_pillar -- host = %s", host)

    if "username" not in kwargs:
        log.error(
            "VMWare external pillar requested but username is not specified in"
            " ext_pillar configuration."
        )
        return vmware_pillar
    else:
        username = kwargs["username"]
        log.debug("vmware_pillar -- username = %s", username)

    if "password" not in kwargs:
        log.error(
            "VMWare external pillar requested but password is not specified in"
            " ext_pillar configuration."
        )
        return vmware_pillar
    else:
        password = kwargs["password"]
        log.debug("vmware_pillar -- password = %s", password)

    if "replace_default_attributes" in kwargs:
        replace_default_attributes = kwargs["replace_default_attributes"]
        if replace_default_attributes:
            pillar_attributes = []
            type_specific_pillar_attributes = {}

    if "property_types" in kwargs:
        for prop_type in kwargs["property_types"]:
            if isinstance(prop_type, dict):
                next_prop_type_key = next(iter(prop_type))
                property_types.append(getattr(vim, next_prop_type_key))
                if isinstance(prop_type[next_prop_type_key], list):
                    pillar_attributes = (
                        pillar_attributes + prop_type[next_prop_type_key]
                    )
                else:
                    log.warning(
                        "A property_type dict was specified, but its value is not a"
                        " list"
                    )
            else:
                property_types.append(getattr(vim, prop_type))
    else:
        property_types = [vim.VirtualMachine]
    log.debug("vmware_pillar -- property_types = %s", property_types)

    if "property_name" in kwargs:
        property_name = kwargs["property_name"]
    else:
        property_name = "name"
    log.debug("vmware_pillar -- property_name = %s", property_name)

    if "protocol" in kwargs:
        protocol = kwargs["protocol"]
        log.debug("vmware_pillar -- protocol = %s", protocol)

    if "port" in kwargs:
        port = kwargs["port"]
        log.debug("vmware_pillar -- port = %s", port)

    virtualgrain = None
    osgrain = None
    if "virtual" in __grains__:
        virtualgrain = __grains__["virtual"].lower()
    if "os" in __grains__:
        osgrain = __grains__["os"].lower()

    if virtualgrain == "vmware" or osgrain == "vmware esxi" or osgrain == "esxi":
        vmware_pillar[pillar_key] = {}
        try:
            _conn = salt.utils.vmware.get_service_instance(
                host,
                username,
                password,
                protocol,
                port,
                verify_ssl=kwargs.get("verify_ssl", True),
            )
            if _conn:
                data = None
                for prop_type in property_types:
                    data = salt.utils.vmware.get_mor_by_property(
                        _conn, prop_type, minion_id, property_name=property_name
                    )
                    if data:
                        type_name = type(data).__name__.replace("vim.", "")
                        if hasattr(data, "availableField"):
                            vmware_pillar[pillar_key]["annotations"] = {}
                            for availableField in data.availableField:
                                for customValue in data.customValue:
                                    if availableField.key == customValue.key:
                                        vmware_pillar[pillar_key]["annotations"][
                                            availableField.name
                                        ] = customValue.value
                        type_specific_pillar_attribute = []
                        if type_name in type_specific_pillar_attributes:
                            type_specific_pillar_attribute = (
                                type_specific_pillar_attributes[type_name]
                            )
                        vmware_pillar[pillar_key] = dictupdate.update(
                            vmware_pillar[pillar_key],
                            _crawl_attribute(
                                data, pillar_attributes + type_specific_pillar_attribute
                            ),
                        )
                        break
                # explicitly disconnect from vCenter when we are done, connections linger idle otherwise
                Disconnect(_conn)
            else:
                log.error(
                    "Unable to obtain a connection with %s, please verify "
                    "your vmware ext_pillar configuration",
                    host,
                )
        except RuntimeError:
            log.error(
                "A runtime error occurred in the vmware_pillar, "
                "this is likely caused by an infinite recursion in "
                "a requested attribute.  Verify your requested attributes "
                "and reconfigure the pillar."
            )

        return vmware_pillar
    else:
        return {}


def _recurse_config_to_dict(t_data):
    """
    helper function to recurse through a vim object and attempt to return all child objects
    """
    if not isinstance(t_data, type(None)):
        if isinstance(t_data, list):
            t_list = []
            for i in t_data:
                t_list.append(_recurse_config_to_dict(i))
            return t_list
        elif isinstance(t_data, dict):
            t_dict = {}
            for k, v in t_data.items():
                t_dict[k] = _recurse_config_to_dict(v)
            return t_dict
        else:
            if hasattr(t_data, "__dict__"):
                return _recurse_config_to_dict(t_data.__dict__)
            else:
                return _serializer(t_data)


def _crawl_attribute(this_data, this_attr):
    """
    helper function to crawl an attribute specified for retrieval
    """
    if isinstance(this_data, list):
        t_list = []
        for d in this_data:
            t_list.append(_crawl_attribute(d, this_attr))
        return t_list
    else:
        if isinstance(this_attr, dict):
            t_dict = {}
            for k in this_attr:
                if hasattr(this_data, k):
                    t_dict[k] = _crawl_attribute(
                        getattr(this_data, k, None), this_attr[k]
                    )
            return t_dict
        elif isinstance(this_attr, list):
            this_dict = {}
            for l in this_attr:
                this_dict = dictupdate.update(this_dict, _crawl_attribute(this_data, l))
            return this_dict
        else:
            return {
                this_attr: _recurse_config_to_dict(getattr(this_data, this_attr, None))
            }


def _serializer(obj):
    """
    helper function to serialize some objects for prettier return
    """
    import datetime

    if isinstance(obj, datetime.datetime):
        if obj.utcoffset() is not None:
            obj = obj - obj.utcoffset()
        return obj.__str__()
    return obj



import pytest

from salt.pillar import extra_minion_data_in_pillar
from tests.support.mock import MagicMock


@pytest.fixture
def configure_loader_modules():
    return {extra_minion_data_in_pillar: {}}


@pytest.fixture
def extra_minion_data():
    return {
        "key1": {"subkey1": "value1"},
        "key2": {"subkey2": {"subsubkey2": "value2"}},
        "key3": "value3",
        "key4": {"subkey4": "value4"},
    }


def test_extra_values_none_or_empty():
    ret = extra_minion_data_in_pillar.ext_pillar(
        "fake_id", MagicMock(), "fake_include", None
    )
    assert ret == {}
    ret = extra_minion_data_in_pillar.ext_pillar(
        "fake_id", MagicMock(), "fake_include", {}
    )
    assert ret == {}


def test_include_all(extra_minion_data):
    for include_all in ["*", "<all>"]:
        ret = extra_minion_data_in_pillar.ext_pillar(
            "fake_id", MagicMock(), include_all, extra_minion_data
        )
        assert ret == extra_minion_data


def test_include_specific_keys(extra_minion_data):
    # Tests partially existing key, key with and without subkey,
    ret = extra_minion_data_in_pillar.ext_pillar(
        "fake_id",
        MagicMock(),
        include=["key1:subkey1", "key2:subkey3", "key3", "key4"],
        extra_minion_data=extra_minion_data,
    )
    assert ret == {
        "key1": {"subkey1": "value1"},
        "key3": "value3",
        "key4": {"subkey4": "value4"},
    }



"""
    Test cases for salt.modules.config
"""


import fnmatch

import pytest

import salt.modules.config as config
from tests.support.mock import patch


@pytest.fixture
def defaults():
    return {
        "test.option.foo": "value of test.option.foo in defaults",
        "test.option.bar": "value of test.option.bar in defaults",
        "test.option.baz": "value of test.option.baz in defaults",
        "test.option": "value of test.option in defaults",
    }


@pytest.fixture
def no_match():
    return "test.option.nope"


@pytest.fixture
def opt_name():
    return "test.option.foo"


@pytest.fixture
def wildcard_opt_name():
    return "test.option.b*"


@pytest.fixture
def configure_loader_modules():
    return {
        config: {
            "__opts__": {
                "test.option.foo": "value of test.option.foo in __opts__",
                "test.option.bar": "value of test.option.bar in __opts__",
                "test.option.baz": "value of test.option.baz in __opts__",
            },
            "__pillar__": {
                "test.option.foo": "value of test.option.foo in __pillar__",
                "test.option.bar": "value of test.option.bar in __pillar__",
                "test.option.baz": "value of test.option.baz in __pillar__",
                "master": {
                    "test.option.foo": "value of test.option.foo in master",
                    "test.option.bar": "value of test.option.bar in master",
                    "test.option.baz": "value of test.option.baz in master",
                },
            },
            "__grains__": {
                "test.option.foo": "value of test.option.foo in __grains__",
                "test.option.bar": "value of test.option.bar in __grains__",
                "test.option.baz": "value of test.option.baz in __grains__",
            },
        }
    }


def _wildcard_match(data, wildcard_opt_name):
    return {x: data[x] for x in fnmatch.filter(data, wildcard_opt_name)}


def test_defaults_only_name(defaults):
    with patch.dict(config.DEFAULTS, defaults):
        opt_name = "test.option"
        opt = config.option(opt_name)
        assert opt == config.DEFAULTS[opt_name]


def test_no_match(defaults, no_match, wildcard_opt_name):
    """
    Make sure that the defa
    """
    with patch.dict(config.DEFAULTS, defaults):
        ret = config.option(no_match)
        assert ret == "", ret

        default = "wat"
        ret = config.option(no_match, default=default)
        assert ret == default, ret

        ret = config.option(no_match, wildcard=True)
        assert ret == {}, ret

        default = {"foo": "bar"}
        ret = config.option(no_match, default=default, wildcard=True)
        assert ret == default, ret

        # Should be no match since wildcard=False
        ret = config.option(wildcard_opt_name)
        assert ret == "", ret


def test_omits(defaults, opt_name, wildcard_opt_name):
    with patch.dict(config.DEFAULTS, defaults):

        # ********** OMIT NOTHING **********

        # Match should be in __opts__ dict
        ret = config.option(opt_name)
        assert ret == config.__opts__[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, wildcard=True)
        assert ret == _wildcard_match(config.__opts__, wildcard_opt_name), ret

        # ********** OMIT __opts__ **********

        # Match should be in __grains__ dict
        ret = config.option(opt_name, omit_opts=True)
        assert ret == config.__grains__[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, omit_opts=True, wildcard=True)
        assert ret == _wildcard_match(config.__grains__, wildcard_opt_name), ret

        # ********** OMIT __opts__, __grains__ **********

        # Match should be in __pillar__ dict
        ret = config.option(opt_name, omit_opts=True, omit_grains=True)
        assert ret == config.__pillar__[opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name, omit_opts=True, omit_grains=True, wildcard=True
        )
        assert ret == _wildcard_match(config.__pillar__, wildcard_opt_name), ret

        # ********** OMIT __opts__, __grains__, __pillar__ **********

        # Match should be in master opts
        ret = config.option(
            opt_name, omit_opts=True, omit_grains=True, omit_pillar=True
        )
        assert ret == config.__pillar__["master"][opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            wildcard=True,
        )
        assert ret == _wildcard_match(
            config.__pillar__["master"], wildcard_opt_name
        ), ret

        # ********** OMIT ALL THE THINGS **********

        # Match should be in master opts
        ret = config.option(
            opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            omit_master=True,
        )
        assert ret == config.DEFAULTS[opt_name], ret

        # Wildcard match
        ret = config.option(
            wildcard_opt_name,
            omit_opts=True,
            omit_grains=True,
            omit_pillar=True,
            omit_master=True,
            wildcard=True,
        )
        assert ret == _wildcard_match(config.DEFAULTS, wildcard_opt_name), ret

        # Match should be in master opts
        ret = config.option(opt_name, omit_all=True)
        assert ret == config.DEFAULTS[opt_name], ret

        # Wildcard match
        ret = config.option(wildcard_opt_name, omit_all=True, wildcard=True)
        assert ret == _wildcard_match(config.DEFAULTS, wildcard_opt_name), ret



"""
Simple Smoke Tests for Connected SSH minions
"""

import pytest
from saltfactories.utils.functional import StateResult

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.mark.skip_if_not_root
def test_service(salt_ssh_cli, grains):
    service = "cron"
    os_family = grains["os_family"]
    os_release = grains["osrelease"]
    if os_family == "RedHat":
        service = "crond"
    elif os_family == "Arch":
        service = "sshd"
    elif os_family == "MacOS":
        service = "org.ntp.ntpd"
        if int(os_release.split(".")[1]) >= 13:
            service = "com.apple.AirPlayXPCHelper"
    ret = salt_ssh_cli.run("service.enable", service)
    assert ret.returncode == 0
    ret = salt_ssh_cli.run("service.get_all")
    assert ret.returncode == 0
    assert service in ret.data
    ret = salt_ssh_cli.run("service.stop", service)
    assert ret.returncode == 0
    ret = salt_ssh_cli.run("service.status", service)
    assert ret.returncode == 0
    assert not ret.data
    ret = salt_ssh_cli.run("service.start", service)
    assert ret.returncode == 0
    ret = salt_ssh_cli.run("service.status", service)
    assert ret.returncode == 0
    assert ret.data


@pytest.fixture
def _state_tree(salt_master, tmp_path):
    top_sls = """
    base:
      '*':
        - core
    """
    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        tmp_path
    )
    with salt_master.state_tree.base.temp_file(
        "top.sls", top_sls
    ), salt_master.state_tree.base.temp_file("core.sls", core_state):
        yield


@pytest.mark.usefixtures("_state_tree")
def test_state_apply(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.apply", "core")
    assert ret.returncode == 0
    state_result = StateResult(ret.data)
    assert state_result.result is True


@pytest.mark.usefixtures("_state_tree")
def test_state_highstate(salt_ssh_cli):
    ret = salt_ssh_cli.run("state.highstate")
    assert ret.returncode == 0
    state_result = StateResult(ret.data)
    assert state_result.result is True



import pytest

import salt.modules.baredoc as baredoc
from tests.support.paths import SALT_CODE_DIR


@pytest.fixture
def configure_loader_modules():
    return {
        baredoc: {
            "__opts__": {"extension_modules": SALT_CODE_DIR},
            "__grains__": {"saltpath": SALT_CODE_DIR},
        }
    }


def test_baredoc_list_states():
    """
    Test baredoc state module listing
    """
    ret = baredoc.list_states(names_only=True)
    assert "value_present" in ret["xml"][0]


def test_baredoc_list_states_args():
    """
    Test baredoc state listing with args
    """
    ret = baredoc.list_states()
    assert "value_present" in ret["xml"][0]
    assert "xpath" in ret["xml"][0]["value_present"]


def test_baredoc_list_states_single():
    """
    Test baredoc state listing single state module
    """
    ret = baredoc.list_states("xml")
    assert "value_present" in ret["xml"][0]
    assert "xpath" in ret["xml"][0]["value_present"]


def test_baredoc_list_modules():
    """
    test baredoc executiion module listing
    """
    ret = baredoc.list_modules(names_only=True)
    assert "get_value" in ret["xml"][0]


def test_baredoc_list_modules_args():
    """
    test baredoc execution module listing with args
    """
    ret = baredoc.list_modules()
    assert "get_value" in ret["xml"][0]
    assert "file" in ret["xml"][0]["get_value"]


def test_baredoc_list_modules_single_and_alias():
    """
    test baredoc single module listing
    """
    ret = baredoc.list_modules("mdata")
    assert "put" in ret["mdata"][2]
    assert "keyname" in ret["mdata"][2]["put"]


def test_baredoc_state_docs():
    ret = baredoc.state_docs()
    assert "XML Manager" in ret["xml"]
    assert "zabbix_usergroup" in ret


def test_baredoc_state_docs_single_arg():
    ret = baredoc.state_docs("xml")
    assert "XML Manager" in ret["xml"]
    ret = baredoc.state_docs("xml.value_present")
    assert "Manages a given XML file" in ret["xml.value_present"]


def test_baredoc_state_docs_multiple_args():
    ret = baredoc.state_docs("zabbix_hostgroup.present", "xml")
    assert "Ensures that the host group exists" in ret["zabbix_hostgroup.present"]
    assert "XML Manager" in ret["xml"]
    assert "Manages a given XML file" in ret["xml.value_present"]


def test_baredoc_module_docs():
    ret = baredoc.module_docs()
    assert "A module for testing" in ret["saltcheck"]


def test_baredoc_module_docs_single_arg():
    ret = baredoc.module_docs("saltcheck")
    assert "A module for testing" in ret["saltcheck"]


def test_baredoc_module_docs_multiple_args():
    ret = baredoc.module_docs("saltcheck", "xml.get_value")
    assert "A module for testing" in ret["saltcheck"]
    assert "Returns the value of the matched xpath element" in ret["xml.get_value"]



"""
Unit tests for file.line
"""

import logging
import os
import shutil

import pytest

import salt.config
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.config as configmod
import salt.modules.file as filemod
import salt.utils.data
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError
from tests.support.mock import DEFAULT, MagicMock, mock_open, patch

log = logging.getLogger(__name__)


class DummyStat:
    st_mode = 33188
    st_ino = 115331251
    st_dev = 44
    st_nlink = 1
    st_uid = 99200001
    st_gid = 99200001
    st_size = 41743
    st_atime = 1552661253
    st_mtime = 1552661253
    st_ctime = 1552661253


@pytest.fixture
def anyattr():
    class AnyAttr:
        def __getattr__(self, item):
            return 0

        def __call__(self, *args, **kwargs):
            return self

    return AnyAttr()


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__salt__": {
                "config.manage_mode": configmod.manage_mode,
                "cmd.run": cmdmod.run,
                "cmd.run_all": cmdmod.run_all,
            },
            "__opts__": {
                "test": False,
                "file_roots": {"base": "tmp"},
                "pillar_roots": {"base": "tmp"},
                "cachedir": "tmp",
                "grains": {},
            },
            "__grains__": {"kernel": "Linux"},
            "__utils__": {"stringutils.get_diff": salt.utils.stringutils.get_diff},
        }
    }


@pytest.fixture
def get_body():
    def _get_body(content):
        """
        The body is written as bytestrings or strings depending on platform.
        This func accepts a string of content and returns the appropriate list
        of strings back.
        """
        ret = content.splitlines(True)
        return salt.utils.data.decode_list(ret, to_str=True)

    return _get_body


@pytest.fixture
def tempfile_name(tmp_path):
    subdir = tmp_path / "file-line-temp-dir"
    subdir.mkdir()
    filename = str(subdir / "file-line-temp-file")

    # File needs to be created
    with salt.utils.files.fopen(filename, "w"):
        pass

    yield filename

    # We need to make sure to remove the tree we just created to avoid clashes with other tests
    shutil.rmtree(str(subdir))


def test_set_line_should_raise_command_execution_error_with_no_mode():
    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(lines=[], mode=None)
    assert str(err.value) == "Mode was not defined. How to process the file?"


def test_set_line_should_raise_command_execution_error_with_unknown_mode():
    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(lines=[], mode="fnord")
    assert str(err.value) == "Unknown mode: fnord"


@pytest.mark.parametrize("mode", ("insert", "ensure", "replace"))
def test_if_content_is_none_and_mode_is_valid_but_not_delete_it_should_raise_command_execution_error(
    mode,
):
    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(lines=[], mode=mode)
    assert str(err.value) == "Content can only be empty if mode is delete"


@pytest.mark.parametrize("mode", ("delete", "replace"))
def test_if_delete_or_replace_is_called_with_empty_lines_it_should_warn_and_return_empty_body(
    mode,
):
    with patch("salt.modules.file.log.warning", MagicMock()) as fake_warn:
        actual_lines = filemod._set_line(mode=mode, lines=[], content="roscivs")
        assert actual_lines == []
        fake_warn.assert_called_with("Cannot find text to %s. File is empty.", mode)


def test_if_mode_is_delete_and_not_before_after_or_match_then_content_should_be_used_to_delete_line():
    lines = ["foo", "roscivs", "bar"]
    to_remove = "roscivs"
    expected_lines = ["foo", "bar"]

    actual_lines = filemod._set_line(mode="delete", lines=lines, content=to_remove)

    assert actual_lines == expected_lines


def test_if_mode_is_replace_and_not_before_after_or_match_and_content_exists_then_lines_should_not_change():
    original_lines = ["foo", "roscivs", "bar"]
    content = "roscivs"

    actual_lines = filemod._set_line(
        mode="replace", lines=original_lines, content=content
    )

    assert actual_lines == original_lines


def test_if_mode_is_replace_and_match_is_set_then_it_should_replace_the_first_match():
    to_replace = "quuxy"
    replacement = "roscivs"
    original_lines = ["foo", to_replace, "bar"]
    expected_lines = ["foo", replacement, "bar"]

    actual_lines = filemod._set_line(
        mode="replace",
        lines=original_lines,
        content=replacement,
        match=to_replace,
    )

    assert actual_lines == expected_lines


def test_if_mode_is_replace_and_indent_is_true_then_it_should_match_indention_of_existing_line():
    indents = "\t\t      \t \t"
    to_replace = indents + "quuxy"
    replacement = "roscivs"
    original_lines = ["foo", to_replace, "bar"]
    expected_lines = ["foo", indents + replacement, "bar"]

    actual_lines = filemod._set_line(
        mode="replace",
        lines=original_lines,
        content=replacement,
        match=to_replace,
        indent=True,
    )

    assert actual_lines == expected_lines


def test_if_mode_is_replace_and_indent_is_false_then_it_should_just_use_content():
    indents = "\t\t      \t \t"
    to_replace = indents + "quuxy"
    replacement = "\t        \t\troscivs"
    original_lines = ["foo", to_replace, "bar"]
    expected_lines = ["foo", replacement, "bar"]

    actual_lines = filemod._set_line(
        mode="replace",
        lines=original_lines,
        content=replacement,
        match=to_replace,
        indent=False,
    )

    assert actual_lines == expected_lines


def test_if_mode_is_insert_and_no_location_before_or_after_then_it_should_raise_command_execution_error():
    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=[],
            content="fnord",
            mode="insert",
            location=None,
            before=None,
            after=None,
        )

    assert (
        str(err.value)
        == 'On insert either "location" or "before/after" conditions are required.'
    )


def test_if_mode_is_insert_and_location_is_start_it_should_insert_content_at_start():
    lines = ["foo", "bar", "bang"]
    content = "roscivs"
    expected_lines = [content] + lines

    with patch("os.linesep", ""):
        actual_lines = filemod._set_line(
            lines=lines,
            content=content,
            mode="insert",
            location="start",
        )

    assert actual_lines == expected_lines


def test_if_mode_is_insert_and_lines_have_eol_then_inserted_line_should_have_matching_eol():
    linesep = "\r\n"
    lines = ["foo" + linesep]
    content = "roscivs"
    expected_lines = [content + linesep] + lines

    actual_lines = filemod._set_line(
        lines=lines,
        content=content,
        mode="insert",
        location="start",
    )

    assert actual_lines == expected_lines


def test_if_mode_is_insert_and_no_lines_then_the_content_should_have_os_linesep_added():
    content = "roscivs"
    fake_linesep = "\U0001f40d"
    expected_lines = [content + fake_linesep]

    with patch("os.linesep", fake_linesep):
        actual_lines = filemod._set_line(
            lines=[],
            content=content,
            mode="insert",
            location="start",
        )

    assert actual_lines == expected_lines


def test_if_location_is_end_of_empty_file_then_it_should_just_be_content():
    content = "roscivs"
    expected_lines = [content]

    actual_lines = filemod._set_line(
        lines=[],
        content=content,
        mode="insert",
        location="end",
    )

    assert actual_lines == expected_lines


def test_if_location_is_end_of_file_and_indent_is_True_then_line_should_match_previous_indent():
    content = "roscivs"
    indent = "   \t\t\t   "
    original_lines = [indent + "fnord"]
    expected_lines = original_lines + [indent + content]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location="end",
        indent=True,
    )

    assert actual_lines == expected_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_if_location_is_not_set_but_before_and_after_are_then_line_should_appear_as_the_line_before_before(
    indent,
):
    content = "roscivs"
    after = "after"
    before = "before"
    original_lines = ["foo", "bar", indent + after, "belowme", indent + before]
    expected_lines = [
        "foo",
        "bar",
        indent + after,
        "belowme",
        indent + content,
        indent + before,
    ]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=None,
        before=before,
        after=after,
    )

    assert actual_lines == expected_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_insert_with_after_and_before_with_no_location_should_indent_to_match_before_indent(
    indent,
):
    content = "roscivs"
    after = "after"
    before = "before"
    original_lines = [
        "foo",
        "bar",
        indent + after,
        "belowme",
        (indent * 2) + before,
    ]
    expected_lines = [
        "foo",
        "bar",
        indent + after,
        "belowme",
        (indent * 2) + content,
        (indent * 2) + before,
    ]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=None,
        before=before,
        after=after,
    )

    assert actual_lines == expected_lines


def test_if_not_location_but_before_and_after_and_more_than_one_after_it_should_CommandExecutionError():
    after = "one"
    before = "two"
    original_lines = [after, after, after, after, before]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=None,
            before=before,
            after=after,
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "after" expression'
    )


def test_if_not_location_but_before_and_after_and_more_than_one_before_it_should_CommandExecutionError():
    after = "one"
    before = "two"
    original_lines = [after, before, before, before]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=None,
            before=before,
            after=after,
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "before" expression'
    )


def test_if_not_location_or_before_but_after_and_after_has_more_than_one_it_should_CommandExecutionError():
    location = None
    before = None
    after = "after"
    original_lines = [after, after, after]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=location,
            before=before,
            after=after,
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "after" expression'
    )


def test_if_not_location_or_after_but_before_and_before_has_more_than_one_it_should_CommandExecutionError():
    location = None
    before = "before"
    after = None
    original_lines = [before, before, before]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=location,
            before=before,
            after=after,
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "before" expression'
    )


def test_if_not_location_or_after_and_no_before_in_lines_it_should_CommandExecutionError():
    location = None
    before = "before"
    after = None
    original_lines = ["fnord", "fnord"]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=location,
            before=before,
            after=after,
        )
    assert str(err.value) == "Neither before or after was found in file"


def test_if_not_location_or_before_and_no_after_in_lines_it_should_CommandExecutionError():
    location = None
    before = None
    after = "after"
    original_lines = ["fnord", "fnord"]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content="fnord",
            mode="insert",
            location=location,
            before=before,
            after=after,
        )
    assert str(err.value) == "Neither before or after was found in file"


def test_if_not_location_or_before_but_after_then_line_should_be_inserted_after_after():
    location = before = None
    after = "indessed"
    content = "roscivs"
    indent = "\t\t\t   "
    original_lines = ["foo", indent + after, "bar"]
    expected_lines = ["foo", indent + after, indent + content, "bar"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
    )

    assert actual_lines == expected_lines


def test_insert_with_after_should_ignore_line_endings_on_comparison():
    after = "after"
    content = "roscivs"
    line_endings = "\r\n\r\n"
    original_lines = [after, content + line_endings]

    actual_lines = filemod._set_line(
        lines=original_lines[:],
        content=content,
        mode="insert",
        after=after,
    )

    assert actual_lines == original_lines


def test_insert_with_before_should_ignore_line_endings_on_comparison():
    before = "before"
    content = "bottia"
    line_endings = "\r\n\r\n"
    original_lines = [content + line_endings, before]

    actual_lines = filemod._set_line(
        lines=original_lines[:],
        content=content,
        mode="insert",
        before=before,
    )

    assert actual_lines == original_lines


def test_if_not_location_or_before_but_after_and_indent_False_then_line_should_be_inserted_after_after_without_indent():
    location = before = None
    after = "indessed"
    content = "roscivs"
    indent = "\t\t\t   "
    original_lines = ["foo", indent + after, "bar"]
    expected_lines = ["foo", indent + after, content, "bar"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
        indent=False,
    )

    assert actual_lines == expected_lines


def test_if_not_location_or_after_but_before_then_line_should_be_inserted_before_before():
    location = after = None
    before = "indessed"
    content = "roscivs"
    indent = "\t\t\t   "
    original_lines = [indent + "foo", indent + before, "bar"]
    expected_lines = [indent + "foo", indent + content, indent + before, "bar"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
    )

    assert actual_lines == expected_lines


def test_if_not_location_or_after_but_before_and_indent_False_then_line_should_be_inserted_before_before_without_indent():
    location = after = None
    before = "indessed"
    content = "roscivs"
    indent = "\t\t\t   "
    original_lines = [indent + "foo", before, "bar"]
    expected_lines = [indent + "foo", content, before, "bar"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
        indent=False,
    )

    assert actual_lines == expected_lines


def test_insert_after_the_last_line_should_work():
    location = before = None
    after = "indessed"
    content = "roscivs"
    original_lines = [after]
    expected_lines = [after, content]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
        indent=True,
    )

    assert actual_lines == expected_lines


def test_insert_should_work_just_like_ensure_on_before():
    # I'm pretty sure that this is or should be a bug, but that
    # is how things currently work, so I'm calling it out here.
    #
    # If that should change, then this test should change.
    before = "indessed"
    content = "roscivs"
    original_lines = [content, before]

    actual_lines = filemod._set_line(
        lines=original_lines[:],
        content=content,
        mode="insert",
        before=before,
    )

    assert actual_lines == original_lines


def test_insert_should_work_just_like_ensure_on_after():
    # I'm pretty sure that this is or should be a bug, but that
    # is how things currently work, so I'm calling it out here.
    #
    # If that should change, then this test should change.
    after = "indessed"
    content = "roscivs"
    original_lines = [after, content]

    actual_lines = filemod._set_line(
        # If we don't pass in a copy of the lines then it modifies
        # them, and our test fails. Oops.
        lines=original_lines[:],
        content=content,
        mode="insert",
        after=after,
    )

    assert actual_lines == original_lines


def test_insert_before_the_first_line_should_work():
    location = after = None
    before = "indessed"
    content = "roscivs"
    original_lines = [before]
    expected_lines = [content, before]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="insert",
        location=location,
        before=before,
        after=after,
        indent=True,
    )

    assert actual_lines == expected_lines


def test_ensure_with_before_and_too_many_after_should_CommandExecutionError():
    location = None
    before = "before"
    after = "after"
    lines = [after, after, before]
    content = "fnord"

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=lines,
            content=content,
            mode="ensure",
            location=location,
            before=before,
            after=after,
        )

    assert (
        str(err.value) == 'Found more than expected occurrences in "after" expression'
    )


def test_ensure_with_too_many_after_should_CommandExecutionError():
    after = "fnord"
    bad_lines = [after, after]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=bad_lines,
            content="asdf",
            after=after,
            mode="ensure",
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "after" expression'
    )


def test_ensure_with_after_and_too_many_before_should_CommandExecutionError():
    location = None
    before = "before"
    after = "after"
    lines = [after, before, before]
    content = "fnord"

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=lines,
            content=content,
            mode="ensure",
            location=location,
            before=before,
            after=after,
        )

    assert (
        str(err.value) == 'Found more than expected occurrences in "before" expression'
    )


def test_ensure_with_too_many_before_should_CommandExecutionError():
    before = "fnord"
    bad_lines = [before, before]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=bad_lines,
            content="asdf",
            before=before,
            mode="ensure",
        )
    assert (
        str(err.value) == 'Found more than expected occurrences in "before" expression'
    )


def test_ensure_with_before_and_after_that_already_contains_the_line_should_return_original_info():
    before = "before"
    after = "after"
    content = "roscivs"
    original_lines = [after, content, before]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        mode="ensure",
        after=after,
        before=before,
    )

    assert actual_lines == original_lines


def test_ensure_with_too_many_lines_between_before_and_after_should_CommandExecutionError():
    before = "before"
    after = "after"
    content = "roscivs"
    original_lines = [after, "fnord", "fnord", before]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=original_lines,
            content=content,
            mode="ensure",
            after=after,
            before=before,
        )

    assert (
        str(err.value)
        == 'Found more than one line between boundaries "before" and "after".'
    )


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_no_lines_between_before_and_after_should_insert_a_line(indent):
    before = "before"
    after = "after"
    content = "roscivs"
    original_lines = [indent + after, before]
    expected_lines = [indent + after, indent + content, before]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == expected_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_existing_but_different_line_should_set_the_line(indent):
    before = "before"
    after = "after"
    content = "roscivs"
    original_lines = [indent + after, "fnord", before]
    expected_lines = [indent + after, indent + content, before]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == expected_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_after_and_existing_content_should_return_same_lines(indent):
    before = None
    after = "after"
    content = "roscivs"
    original_lines = [indent + after, indent + content, "fnord"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == original_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_after_and_missing_content_should_add_it(indent):
    before = None
    after = "after"
    content = "roscivs"
    original_lines = [indent + after, "more fnord", "fnord"]
    expected_lines = [indent + after, indent + content, "more fnord", "fnord"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == expected_lines


def test_ensure_with_after_and_content_at_the_end_should_not_add_duplicate():
    after = "after"
    content = "roscivs"
    original_lines = [after, content + "\n"]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        after=after,
        mode="ensure",
    )

    assert actual_lines == original_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_before_and_missing_content_should_add_it(indent):
    before = "before"
    after = None
    content = "roscivs"
    original_lines = [indent + "fnord", indent + "fnord", before]
    expected_lines = [
        indent + "fnord",
        indent + "fnord",
        indent + content,
        before,
    ]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == expected_lines


@pytest.mark.parametrize("indent", ("", " \t \t\t\t      "))
def test_ensure_with_before_and_existing_content_should_return_same_lines(indent):
    before = "before"
    after = None
    content = "roscivs"
    original_lines = [indent + "fnord", indent + content, before]

    actual_lines = filemod._set_line(
        lines=original_lines,
        content=content,
        before=before,
        after=after,
        mode="ensure",
        indent=True,
    )

    assert actual_lines == original_lines


def test_ensure_without_before_and_after_should_CommandExecutionError():
    before = "before"
    after = "after"
    bad_lines = ["fnord", "fnord1", "fnord2"]

    with pytest.raises(CommandExecutionError) as err:
        filemod._set_line(
            lines=bad_lines,
            before=before,
            after=after,
            content="aardvark",
            mode="ensure",
        )
    assert (
        str(err.value)
        == "Wrong conditions? Unable to ensure line without knowing where"
        " to put it before and/or after."
    )


@pytest.mark.parametrize("mode", ["delete", "replace"])
def test_delete_line_in_empty_file(anyattr, mode):
    """
    Tests that when calling file.line with ``mode=delete``,
    the function doesn't stack trace if the file is empty.
    Should return ``False``.

    See Issue #38438.
    """
    with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ):
        _log = MagicMock()
        with patch("salt.utils.files.fopen", mock_open(read_data="")), patch(
            "os.stat", anyattr
        ), patch("salt.modules.file.log", _log):
            assert not filemod.line(
                "/dummy/path", content="foo", match="bar", mode=mode
            )
        warning_call = _log.warning.call_args_list[0][0]
        warning_log_msg = warning_call[0] % warning_call[1:]
        assert "Cannot find text to {}".format(mode) in warning_log_msg


@pytest.mark.parametrize("mode", ["delete", "replace"])
def test_line_delete_no_match(mode):
    """
    Tests that when calling file.line with ``mode=delete``,
    with not matching pattern to delete returns False
    :return:
    """
    file_content = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/custom"]
    )
    match = "not matching"
    with patch("os.path.realpath", MagicMock()), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ), patch("os.stat", MagicMock()):
        files_fopen = mock_open(read_data=file_content)
        with patch("salt.utils.files.fopen", files_fopen):
            atomic_opener = mock_open()
            with patch("salt.utils.atomicfile.atomic_open", atomic_opener):
                assert not filemod.line("foo", content="foo", match=match, mode=mode)


@pytest.mark.parametrize(
    "mode,err_msg",
    [
        (None, "How to process the file"),
        ("nonsense", "Unknown mode"),
    ],
)
def test_line_modecheck_failure(mode, err_msg):
    """
    Test for file.line for empty or wrong mode.
    Calls unknown or empty mode and expects failure.
    :return:
    """
    with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ):
        with pytest.raises(CommandExecutionError) as exc_info:
            filemod.line("foo", mode=mode)
        assert err_msg in str(exc_info.value)


@pytest.mark.parametrize("mode", ["insert", "ensure", "replace"])
def test_line_no_content(mode):
    """
    Test for file.line for an empty content when not deleting anything.
    :return:
    """
    with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ):
        with pytest.raises(CommandExecutionError) as exc_info:
            filemod.line("foo", mode=mode)
        assert 'Content can only be empty if mode is "delete"' in str(exc_info.value)


def test_line_insert_no_location_no_before_no_after():
    """
    Test for file.line for insertion but define no location/before/after.
    :return:
    """
    with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ), patch("os.stat", MagicMock()), patch(
        "salt.utils.files.fopen", mock_open(read_data="test data")
    ):
        with pytest.raises(CommandExecutionError) as exc_info:
            filemod.line("foo", content="test content", mode="insert")
        assert '"location" or "before/after"' in str(exc_info.value)


def test_line_insert_after_no_pattern(tempfile_name, get_body):
    """
    Test for file.line for insertion after specific line, using no pattern.

    See issue #38670
    :return:
    """
    file_content = os.linesep.join(["file_roots:", "  base:", "    - /srv/salt"])
    file_modified = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/custom"]
    )
    cfg_content = "- /srv/custom"

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name, content=cfg_content, after="- /srv/salt", mode="insert"
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


@pytest.mark.parametrize("after_line", ["file_r.*", ".*roots"])
def test_line_insert_after_pattern(tempfile_name, get_body, after_line):
    """
    Test for file.line for insertion after specific line, using pattern.

    See issue #38670
    :return:
    """
    file_content = os.linesep.join(
        [
            "file_boots:",
            "  - /rusty",
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/sugar",
        ]
    )
    file_modified = os.linesep.join(
        [
            "file_boots:",
            "  - /rusty",
            "file_roots:",
            "  custom:",
            "    - /srv/custom",
            "  base:",
            "    - /srv/salt",
            "    - /srv/sugar",
        ]
    )
    cfg_content = os.linesep.join(["  custom:", "    - /srv/custom"])
    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            after=after_line,
            mode="insert",
            indent=False,
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        # We passed cfg_content with a newline in the middle, so it
        # will be written as two lines in the same element of the list
        # passed to .writelines()
        expected[3] = expected[3] + expected.pop(4)
        assert writelines_content[0] == expected, (
            writelines_content[0],
            expected,
        )


def test_line_insert_multi_line_content_after_unicode(tempfile_name, get_body):
    """
    Test for file.line for insertion after specific line with Unicode

    See issue #48113
    :return:
    """
    file_content = "This is a line{}This is another line".format(os.linesep)
    file_modified = salt.utils.stringutils.to_str(
        "This is a line{}"
        "This is another line{}"
        "This is a line with unicode Ŷ".format(os.linesep, os.linesep)
    )
    cfg_content = "This is a line with unicode Ŷ"
    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    after_line = "This is another line"
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            after=after_line,
            mode="insert",
            indent=False,
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (
            writelines_content[0],
            expected,
        )


@pytest.mark.parametrize("before_line", ["/srv/salt", "/srv/sa.*t"])
def test_line_insert_before(tempfile_name, get_body, before_line):
    """
    Test for file.line for insertion before specific line, using pattern and no patterns.

    See issue #38670
    :return:
    """
    file_content = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/sugar"]
    )
    file_modified = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/custom",
            "    - /srv/salt",
            "    - /srv/sugar",
        ]
    )
    cfg_content = "- /srv/custom"

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name, content=cfg_content, before=before_line, mode="insert"
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        # assert writelines_content[0] == expected, (writelines_content[0], expected)
        assert writelines_content[0] == expected


def test_line_assert_exception_pattern():
    """
    Test for file.line for exception on insert with too general pattern.

    :return:
    """
    file_content = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/sugar"]
    )
    cfg_content = "- /srv/custom"
    before_line = "/sr.*"
    with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
        "os.path.isfile", MagicMock(return_value=True)
    ), patch("os.stat", MagicMock()), patch(
        "salt.utils.files.fopen", mock_open(read_data=file_content)
    ), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ):
        with pytest.raises(CommandExecutionError) as cm:
            filemod.line(
                "foo",
                content=cfg_content,
                before=before_line,
                mode="insert",
            )
        assert (
            str(cm.value)
            == 'Found more than expected occurrences in "before" expression'
        )


def test_line_insert_before_after(tempfile_name, get_body):
    """
    Test for file.line for insertion before specific line, using pattern and no patterns.

    See issue #38670
    :return:
    """
    file_content = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/pepper",
            "    - /srv/sugar",
        ]
    )
    file_modified = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/pepper",
            "    - /srv/coriander",
            "    - /srv/sugar",
        ]
    )
    cfg_content = "- /srv/coriander"

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    b_line = "/srv/sugar"
    a_line = "/srv/salt"
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            before=b_line,
            after=a_line,
            mode="insert",
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected


def test_line_insert_start(tempfile_name, get_body):
    """
    Test for file.line for insertion at the beginning of the file
    :return:
    """
    cfg_content = "everything: fantastic"
    file_content = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/sugar"]
    )
    file_modified = os.linesep.join(
        [
            cfg_content,
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/sugar",
        ]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name, content=cfg_content, location="start", mode="insert"
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


def test_line_insert_end(tempfile_name, get_body):
    """
    Test for file.line for insertion at the end of the file (append)
    :return:
    """
    cfg_content = "everything: fantastic"
    file_content = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/sugar"]
    )
    file_modified = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/sugar",
            "    " + cfg_content,
        ]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(tempfile_name, content=cfg_content, location="end", mode="insert")
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


def test_line_insert_ensure_before(tempfile_name, get_body):
    """
    Test for file.line for insertion ensuring the line is before
    :return:
    """
    cfg_content = "/etc/init.d/someservice restart"
    file_content = os.linesep.join(["#!/bin/bash", "", "exit 0"])
    file_modified = os.linesep.join(["#!/bin/bash", "", cfg_content, "exit 0"])

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(tempfile_name, content=cfg_content, before="exit 0", mode="ensure")
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


def test_line_insert_duplicate_ensure_before(tempfile_name):
    """
    Test for file.line for insertion ensuring the line is before
    :return:
    """
    cfg_content = "/etc/init.d/someservice restart"
    file_content = os.linesep.join(["#!/bin/bash", "", cfg_content, "exit 0"])
    file_modified = os.linesep.join(["#!/bin/bash", "", cfg_content, "exit 0"])

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(tempfile_name, content=cfg_content, before="exit 0", mode="ensure")
        # If file not modified no handlers in dict
        assert atomic_open_mock.filehandles.get(tempfile_name) is None


def test_line_insert_ensure_before_first_line(tempfile_name, get_body):
    """
    Test for file.line for insertion ensuring the line is before first line
    :return:
    """
    cfg_content = "#!/bin/bash"
    file_content = os.linesep.join(["/etc/init.d/someservice restart", "exit 0"])
    file_modified = os.linesep.join(
        [cfg_content, "/etc/init.d/someservice restart", "exit 0"]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            before="/etc/init.d/someservice restart",
            mode="ensure",
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


def test_line_insert_ensure_after(tempfile_name, get_body):
    """
    Test for file.line for insertion ensuring the line is after
    :return:
    """
    cfg_content = "exit 0"
    file_content = os.linesep.join(["#!/bin/bash", "/etc/init.d/someservice restart"])
    file_modified = os.linesep.join(
        ["#!/bin/bash", "/etc/init.d/someservice restart", cfg_content]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            after="/etc/init.d/someservice restart",
            mode="ensure",
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (writelines_content[0], expected)


def test_line_insert_duplicate_ensure_after(tempfile_name):
    """
    Test for file.line for insertion ensuring the line is after
    :return:
    """
    cfg_content = "exit 0"
    file_content = os.linesep.join(
        ["#!/bin/bash", "/etc/init.d/someservice restart", cfg_content]
    )
    file_modified = os.linesep.join(
        ["#!/bin/bash", "/etc/init.d/someservice restart", cfg_content]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content=cfg_content,
            after="/etc/init.d/someservice restart",
            mode="ensure",
        )
        # If file not modified no handlers in dict
        assert atomic_open_mock.filehandles.get(tempfile_name) is None


def test_line_insert_ensure_beforeafter_twolines(tempfile_name, get_body):
    """
    Test for file.line for insertion ensuring the line is between two lines
    :return:
    """
    cfg_content = 'EXTRA_GROUPS="dialout cdrom floppy audio video plugdev users"'
    # pylint: disable=W1401
    file_content = os.linesep.join(
        [
            r'NAME_REGEX="^[a-z][-a-z0-9_]*\$"',
            'SKEL_IGNORE_REGEX="dpkg-(old|new|dist|save)"',
        ]
    )
    # pylint: enable=W1401
    after, before = file_content.split(os.linesep)
    file_modified = os.linesep.join([after, cfg_content, before])

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    for (_after, _before) in [(after, before), ("NAME_.*", "SKEL_.*")]:
        with patch("os.path.isfile", isfile_mock), patch(
            "os.stat", MagicMock(return_value=DummyStat())
        ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
            "salt.utils.atomicfile.atomic_open", mock_open()
        ) as atomic_open_mock:
            filemod.line(
                tempfile_name,
                content=cfg_content,
                after=_after,
                before=_before,
                mode="ensure",
            )
            handles = atomic_open_mock.filehandles[tempfile_name]
            # We should only have opened the file once
            open_count = len(handles)
            assert open_count == 1, open_count
            # We should only have invoked .writelines() once...
            writelines_content = handles[0].writelines_calls
            writelines_count = len(writelines_content)
            assert writelines_count == 1, writelines_count
            # ... with the updated content
            expected = get_body(file_modified)
            assert writelines_content[0] == expected, (
                writelines_content[0],
                expected,
            )


def test_line_insert_ensure_beforeafter_twolines_exists(tempfile_name):
    """
    Test for file.line for insertion ensuring the line is between two lines
    where content already exists
    """
    cfg_content = 'EXTRA_GROUPS="dialout"'
    # pylint: disable=W1401
    file_content = os.linesep.join(
        [
            r'NAME_REGEX="^[a-z][-a-z0-9_]*\$"',
            'EXTRA_GROUPS="dialout"',
            'SKEL_IGNORE_REGEX="dpkg-(old|new|dist|save)"',
        ]
    )
    # pylint: enable=W1401
    after, before = (
        file_content.split(os.linesep)[0],
        file_content.split(os.linesep)[2],
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    for (_after, _before) in [(after, before), ("NAME_.*", "SKEL_.*")]:
        with patch("os.path.isfile", isfile_mock), patch(
            "os.stat", MagicMock(return_value=DummyStat())
        ), patch("salt.utils.files.fopen", mock_open(read_data=file_content)), patch(
            "salt.utils.atomicfile.atomic_open", mock_open()
        ) as atomic_open_mock:
            result = filemod.line(
                "foo",
                content=cfg_content,
                after=_after,
                before=_before,
                mode="ensure",
            )
            # We should not have opened the file
            assert not atomic_open_mock.filehandles
            # No changes should have been made
            assert result is False


def test_line_insert_ensure_beforeafter_rangelines():
    """
    Test for file.line for insertion ensuring the line is between two lines
    within the range.  This expected to bring no changes.
    """
    cfg_content = 'EXTRA_GROUPS="dialout cdrom floppy audio video plugdev users"'
    # pylint: disable=W1401
    file_content = (
        r'NAME_REGEX="^[a-z][-a-z0-9_]*\$"{}SETGID_HOME=no{}ADD_EXTRA_GROUPS=1{}'
        'SKEL_IGNORE_REGEX="dpkg-(old|new|dist|save)"'.format(
            os.linesep, os.linesep, os.linesep
        )
    )
    # pylint: enable=W1401
    after, before = (
        file_content.split(os.linesep)[0],
        file_content.split(os.linesep)[-1],
    )
    for (_after, _before) in [(after, before), ("NAME_.*", "SKEL_.*")]:
        with patch("os.path.realpath", MagicMock(wraps=lambda x: x)), patch(
            "os.path.isfile", MagicMock(return_value=True)
        ), patch("os.stat", MagicMock()), patch(
            "salt.utils.files.fopen", mock_open(read_data=file_content)
        ), patch(
            "salt.utils.atomicfile.atomic_open", mock_open()
        ):
            with pytest.raises(CommandExecutionError) as exc_info:
                filemod.line(
                    "foo",
                    content=cfg_content,
                    after=_after,
                    before=_before,
                    mode="ensure",
                )
            assert (
                'Found more than one line between boundaries "before" and "after"'
                in str(exc_info.value)
            )


@pytest.mark.parametrize(
    "content", ["/srv/pepper", "/srv/pepp*", "/srv/p.*", "/sr.*pe.*"]
)
def test_line_delete(tempfile_name, get_body, content):
    """
    Test for file.line for deletion of specific line
    :return:
    """
    file_content = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/pepper",
            "    - /srv/sugar",
        ]
    )
    file_modified = os.linesep.join(
        ["file_roots:", "  base:", "    - /srv/salt", "    - /srv/sugar"]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    files_fopen = mock_open(read_data=file_content)
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", files_fopen), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(tempfile_name, content=content, mode="delete")
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (
            writelines_content[0],
            expected,
        )


@pytest.mark.parametrize(
    "match", ["/srv/pepper", "/srv/pepp*", "/srv/p.*", "/sr.*pe.*"]
)
def test_line_replace(tempfile_name, get_body, match):
    """
    Test for file.line for replacement of specific line
    :return:
    """
    file_content = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/pepper",
            "    - /srv/sugar",
        ]
    )
    file_modified = os.linesep.join(
        [
            "file_roots:",
            "  base:",
            "    - /srv/salt",
            "    - /srv/natrium-chloride",
            "    - /srv/sugar",
        ]
    )

    isfile_mock = MagicMock(
        side_effect=lambda x: True if x == tempfile_name else DEFAULT
    )
    files_fopen = mock_open(read_data=file_content)
    with patch("os.path.isfile", isfile_mock), patch(
        "os.stat", MagicMock(return_value=DummyStat())
    ), patch("salt.utils.files.fopen", files_fopen), patch(
        "salt.utils.atomicfile.atomic_open", mock_open()
    ) as atomic_open_mock:
        filemod.line(
            tempfile_name,
            content="- /srv/natrium-chloride",
            match=match,
            mode="replace",
        )
        handles = atomic_open_mock.filehandles[tempfile_name]
        # We should only have opened the file once
        open_count = len(handles)
        assert open_count == 1, open_count
        # We should only have invoked .writelines() once...
        writelines_content = handles[0].writelines_calls
        writelines_count = len(writelines_content)
        assert writelines_count == 1, writelines_count
        # ... with the updated content
        expected = get_body(file_modified)
        assert writelines_content[0] == expected, (
            writelines_content[0],
            expected,
        )



import pytest

import salt.utils.data
import salt.utils.reactor as reactor
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def minion_reactor(minion_opts):
    return reactor.Reactor(minion_opts)


@pytest.fixture
def master_reactor(master_opts):
    return reactor.Reactor(master_opts)


@pytest.mark.skip_on_windows(reason="Reactors unavailable on Windows")
def test_run_minion_reactor(minion_opts, minion_reactor):
    """
    Ensure that list_reactors() returns the correct list of reactor SLS
    files for each tag.
    """

    with patch.object(salt.utils.event, "get_event", MagicMock()):
        os_nice_mock = MagicMock(return_value=True)
        with patch("os.nice", os_nice_mock):
            # By default os.nice should not be called
            minion_reactor.run()
            assert os_nice_mock.mock_calls == []

            # if reactor_niceness is set, os.nice should be called with that value
            with patch.dict(minion_opts, {"reactor_niceness": 9}):
                minion_reactor.run()
                calls = [call(9)]
                os_nice_mock.assert_has_calls(calls)


@pytest.mark.skip_on_windows(reason="Reactors unavailable on Windows")
def test_run_master_reactor(master_opts, master_reactor):
    """
    Ensure that list_reactors() returns the correct list of reactor SLS
    files for each tag.
    """

    with patch.object(salt.utils.event, "get_event", MagicMock()):
        os_nice_mock = MagicMock(return_value=True)
        with patch("os.nice", os_nice_mock):
            # By default os.nice should not be called
            master_reactor.run()
            assert os_nice_mock.mock_calls == []

            # if reactor_niceness is set, os.nice should be called with that value
            with patch.dict(master_opts, {"reactor_niceness": 9}):
                master_reactor.run()
                calls = [call(9)]
                os_nice_mock.assert_has_calls(calls)



"""
    tests.unit.utils.beacons
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test the salt beacon utils
"""

import logging

import salt.utils.beacons as beacons

log = logging.getLogger(__name__)


def test_list_to_dict():
    """
    Make sure you can instantiate etc.
    """
    ret = beacons.list_to_dict([])
    assert isinstance(ret, dict)

    config = [{"a": "b"}, {"c": "d"}, {"_e": "f"}]
    ret = beacons.list_to_dict(config)
    assert isinstance(ret, dict)


def test_remove_hidden_options():
    """
    Make sure you can instantiate etc.
    """
    config = [{"a": "b"}, {"c": "d"}, {"_e": "f"}]
    expected = [{"a": "b"}, {"c": "d"}]
    ret = beacons.remove_hidden_options(config, [])
    assert ret == expected

    config = [{"a": "b"}, {"c": "d"}, {"_e": "f"}]
    expected = [{"a": "b"}, {"c": "d"}, {"_e": "f"}]
    ret = beacons.remove_hidden_options(config, ["_e"])
    assert ret == expected



import logging
import pathlib

import pytest

import salt.states.ssh_auth as ssh_auth_state
import salt.utils.files

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(modules, minion_opts):
    loader = {"__salt__": modules, "__opts__": minion_opts, "__env__": "base"}
    return {ssh_auth_state: loader}


@pytest.fixture(scope="module")
def system_user():
    with pytest.helpers.create_account() as system_account:
        yield system_account


@pytest.mark.skip_if_not_root
@pytest.mark.destructive_test
@pytest.mark.slow_test
def test_ssh_auth_config(tmp_path, system_user, state_tree):
    """
    test running ssh_auth state when
    different config is set. Ensure
    it does not edit the default config.
    """
    userdetails = system_user.info
    user_ssh_dir = pathlib.Path(userdetails.home, ".ssh")
    ret = ssh_auth_state.manage(
        name="test",
        user=system_user.username,
        ssh_keys=["ssh-dss AAAAB3NzaCL0sQ9fJ5bYTEyY== root@domain"],
    )
    with salt.utils.files.fopen(user_ssh_dir / "authorized_keys") as fp:
        pre_data = fp.read()
    file_contents = "ssh-dss AAAAB3NzaCL0sQ9fJ5bYTEyY== root@domain"
    new_auth_file = tmp_path / "authorized_keys3"
    with pytest.helpers.temp_file("authorized", file_contents, state_tree):
        ssh_auth_state.manage(
            name="test",
            user=system_user.username,
            source=f"salt://authorized",
            config=str(new_auth_file),
            ssh_keys=[""],
        )
    with salt.utils.files.fopen(user_ssh_dir / "authorized_keys") as fp:
        post_data = fp.read()
    assert pre_data == post_data
    with salt.utils.files.fopen(new_auth_file) as fp:
        data = fp.read().strip()
    assert data == file_contents



"""
    :codeauthor: Piter Punk <piterpunk@slackware.com>
"""
import pytest

import salt.states.sysfs as sysfs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {sysfs: {}}


def test_if_the_sysfs_attribute_exists():
    """
    Test sysfs.present for a non-existent attribute
    """
    name = "block/sda/queue/this_does_not_exist"
    value = "none"
    comment = "SysFS attribute {} doesn't exist.".format(name)
    ret = {"name": name, "result": False, "changes": {}, "comment": comment}

    mock_read = MagicMock(return_value=False)
    with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
        assert sysfs.present(name, value) == ret


def test_name_is_an_object_and_not_an_attribute():
    """
    Test sysfs.present targeting an object and not one of its attributes
    """
    name = "block/sda/queue"
    value = "none"
    comment = "{} is not a SysFS attribute.".format(name)
    ret = {"name": name, "result": False, "changes": {}, "comment": comment}

    read_from_sysfs = {
        "rotational": 1,
        "rq_affinity": 1,
        "scheduler": "[none] mq-deadline",
    }

    mock_read = MagicMock(return_value=read_from_sysfs)
    with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
        assert sysfs.present(name, value) == ret


def test_already_set():
    """
    Test sysfs.present with equal old and new values
    """
    name = "block/sda/queue"
    value = "none"
    comment = "SysFS attribute {} is already set.".format(name)
    ret = {"name": name, "result": True, "changes": {}, "comment": comment}

    read_from_sysfs = "[none] mq-deadline"

    mock_read = MagicMock(return_value=read_from_sysfs)
    with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
        assert sysfs.present(name, value) == ret


def test_set_new_value_with_test_equals_true():
    """
    Test sysfs.present setting a new value
    """
    name = "devices/system/cpu/cpufreq/policy0"
    value = "powersave"
    comment = "SysFS attribute {} set to be changed.".format(name)
    ret = {"name": name, "result": None, "changes": {}, "comment": comment}

    read_from_sysfs = "performance"

    mock_read = MagicMock(return_value=read_from_sysfs)
    with patch.dict(sysfs.__opts__, {"test": True}):
        with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
            assert sysfs.present(name, value) == ret


def test_set_new_value_with_success():
    """
    Test sysfs.present setting a new value
    """
    name = "block/sda/queue/scheduler"
    value = "mq-deadline"
    comment = "Updated SysFS attribute {} to {}".format(name, value)
    ret = {"name": name, "result": True, "changes": {name: value}, "comment": comment}

    read_from_sysfs = "[none] mq-deadline"

    mock_read = MagicMock(return_value=read_from_sysfs)
    with patch.dict(sysfs.__opts__, {"test": False}):
        with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
            mock_write = MagicMock(return_value=True)
            with patch.dict(sysfs.__salt__, {"sysfs.write": mock_write}):
                assert sysfs.present(name, value) == ret


def test_set_new_value_with_failure():
    """
    Test sysfs.present failure writing the value
    """
    name = "block/sda/queue/scheduler"
    value = "imaginary_scheduler"
    comment = "Failed to set {} to {}".format(name, value)
    ret = {"name": name, "result": False, "changes": {}, "comment": comment}

    read_from_sysfs = "[none] mq-deadline"

    mock_read = MagicMock(return_value=read_from_sysfs)
    with patch.dict(sysfs.__opts__, {"test": False}):
        with patch.dict(sysfs.__salt__, {"sysfs.read": mock_read}):
            mock_write = MagicMock(return_value=False)
            with patch.dict(sysfs.__salt__, {"sysfs.write": mock_write}):
                assert sysfs.present(name, value) == ret



import logging
import os
import shutil
import subprocess
import time

import pytest
from pytestshellutils.exceptions import FactoryTimeout

import salt.utils.platform

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def salt_mm_master_1(request, salt_factories):
    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
    }

    factory = salt_factories.salt_master_daemon(
        "mm-master-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="package")
def mm_master_1_salt_cli(salt_mm_master_1):
    return salt_mm_master_1.salt_cli(timeout=120)


@pytest.fixture(scope="package")
def salt_mm_master_2(salt_factories, salt_mm_master_1):
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        subprocess.check_output(["ifconfig", "lo0", "alias", "127.0.0.2", "up"])

    config_defaults = {
        "open_mode": True,
        "transport": salt_mm_master_1.config["transport"],
    }
    config_overrides = {
        "interface": "127.0.0.2",
    }

    # Use the same ports for both masters, they are binding to different interfaces
    for key in (
        "ret_port",
        "publish_port",
    ):
        config_overrides[key] = salt_mm_master_1.config[key]
    factory = salt_factories.salt_master_daemon(
        "mm-master-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )

    # The secondary salt master depends on the primarily salt master fixture
    # because we need to clone the keys
    for keyfile in ("master.pem", "master.pub"):
        shutil.copyfile(
            os.path.join(salt_mm_master_1.config["pki_dir"], keyfile),
            os.path.join(factory.config["pki_dir"], keyfile),
        )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="package")
def mm_master_2_salt_cli(salt_mm_master_2):
    return salt_mm_master_2.salt_cli(timeout=120)


@pytest.fixture(scope="package")
def salt_mm_minion_1(salt_mm_master_1, salt_mm_master_2):
    config_defaults = {
        "transport": salt_mm_master_1.config["transport"],
    }

    mm_master_1_port = salt_mm_master_1.config["ret_port"]
    mm_master_1_addr = salt_mm_master_1.config["interface"]
    mm_master_2_port = salt_mm_master_2.config["ret_port"]
    mm_master_2_addr = salt_mm_master_2.config["interface"]
    config_overrides = {
        "master": [
            "{}:{}".format(mm_master_1_addr, mm_master_1_port),
            "{}:{}".format(mm_master_2_addr, mm_master_2_port),
        ],
        "test.foo": "baz",
    }
    factory = salt_mm_master_1.salt_minion_daemon(
        "mm-minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="package")
def salt_mm_minion_2(salt_mm_master_1, salt_mm_master_2):
    config_defaults = {
        "transport": salt_mm_master_1.config["transport"],
    }

    mm_master_1_port = salt_mm_master_1.config["ret_port"]
    mm_master_1_addr = salt_mm_master_1.config["interface"]
    mm_master_2_port = salt_mm_master_2.config["ret_port"]
    mm_master_2_addr = salt_mm_master_2.config["interface"]
    config_overrides = {
        "master": [
            "{}:{}".format(mm_master_1_addr, mm_master_1_port),
            "{}:{}".format(mm_master_2_addr, mm_master_2_port),
        ],
        "test.foo": "baz",
    }
    factory = salt_mm_master_2.salt_minion_daemon(
        "mm-minion-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture(scope="package")
def run_salt_cmds():
    def _run_salt_cmds_fn(clis, minions):
        """
        Run test.ping from all clis to all minions
        """
        returned_minions = []
        minion_instances = {minion.id: minion for minion in minions}
        clis_to_check = {minion.id: list(clis) for minion in minions}

        attempts = 6
        timeout = 5
        if salt.utils.platform.spawning_platform():
            timeout *= 2
        while attempts:
            if not clis_to_check:
                break
            for minion in list(clis_to_check):
                if not clis_to_check[minion]:
                    clis_to_check.pop(minion)
                    continue
                for cli in list(clis_to_check[minion]):
                    try:
                        ret = cli.run(
                            "--timeout={}".format(timeout),
                            "test.ping",
                            minion_tgt=minion,
                            _timeout=2 * timeout,
                        )
                        if ret.returncode == 0 and ret.data is True:
                            returned_minions.append((cli, minion_instances[minion]))
                            clis_to_check[minion].remove(cli)
                    except FactoryTimeout:
                        log.debug(
                            "Failed to execute test.ping from %s to %s.",
                            cli.get_display_name(),
                            minion,
                        )
            time.sleep(1)
            attempts -= 1

        return returned_minions

    return _run_salt_cmds_fn


@pytest.fixture(autouse=True)
def ensure_connections(
    salt_mm_minion_1,
    salt_mm_minion_2,
    mm_master_1_salt_cli,
    mm_master_2_salt_cli,
    run_salt_cmds,
):
    # define the function
    def _ensure_connections_fn(clis, minions):
        retries = 3
        while retries:
            returned = run_salt_cmds(clis, minions)
            if len(returned) == len(clis) * len(minions):
                break
            time.sleep(10)
            retries -= 1
        else:
            pytest.fail("Could not ensure the connections were okay.")

    # run the function to ensure initial connections
    _ensure_connections_fn(
        [mm_master_1_salt_cli, mm_master_2_salt_cli],
        [salt_mm_minion_1, salt_mm_minion_2],
    )

    # Give this function back for further use in test fn bodies
    return _ensure_connections_fn



"""
    Tests for xml module
"""


import pytest

from salt.modules import xml


@pytest.fixture
def xml_string():
    return """
    <root xmlns:foo="http://www.foo.org/" xmlns:bar="http://www.bar.org">
        <actors>
            <actor id="1">Christian Bale</actor>
            <actor id="2">Liam Neeson</actor>
            <actor id="3">Michael Caine</actor>
        </actors>
        <foo:singers>
            <foo:singer id="4">Tom Waits</foo:singer>
            <foo:singer id="5">B.B. King</foo:singer>
            <foo:singer id="6">Ray Charles</foo:singer>
        </foo:singers>
    </root>
    """


@pytest.fixture
def configure_loader_modules():
    return {xml: {}}


def test_get_value(xml_string, tmp_path):
    """
    Verify xml.get_value
    """
    xml_file = tmp_path / "test_xml.xml"
    xml_file.write_text(xml_string)
    xml_result = xml.get_value(str(xml_file), ".//actor[@id='2']")
    assert xml_result == "Liam Neeson"


def test_set_value(xml_string, tmp_path):
    """
    Verify xml.set_value
    """
    xml_file = tmp_path / "test_xml.xml"
    xml_file.write_text(xml_string)
    xml_result = xml.set_value(str(xml_file), ".//actor[@id='2']", "Patrick Stewart")
    assert xml_result is True
    xml_result = xml.get_value(str(xml_file), ".//actor[@id='2']")
    assert xml_result == "Patrick Stewart"


def test_get_attribute(xml_string, tmp_path):
    """
    Verify xml.get_attribute
    """
    xml_file = tmp_path / "test_xml.xml"
    xml_file.write_text(xml_string)
    xml_result = xml.get_attribute(str(xml_file), ".//actor[@id='3']")
    assert xml_result == {"id": "3"}


def test_set_attribute(xml_string, tmp_path):
    """
    Verify xml.set_value
    """
    xml_file = tmp_path / "test_xml.xml"
    xml_file.write_text(xml_string)
    xml_result = xml.set_attribute(
        str(xml_file), ".//actor[@id='3']", "edited", "uh-huh"
    )
    assert xml_result is True
    xml_result = xml.get_attribute(str(xml_file), ".//actor[@id='3']")
    assert xml_result == {"edited": "uh-huh", "id": "3"}



import salt.modules.debian_ip as debian_ip


def test_when_no_adapters_are_passed_to_filter_none_should_be_returned():
    no_adapters = {}

    actual_adapters = debian_ip._filter_malformed_interfaces(adapters=no_adapters)

    assert not actual_adapters


def test_when_adapters_only_contains_source_then_source_should_be_returned():
    souce_only_adapters = {"source": "fnord"}
    expected_adapters = souce_only_adapters.copy()

    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=souce_only_adapters
    )

    assert actual_adapters == expected_adapters


def test_when_data_is_missing_from_adapters_they_should_not_be_in_result():
    missing_data_adapters = {"no data adapter": "no dayta is here"}
    expected_adapters = {}
    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=missing_data_adapters
    )

    assert actual_adapters == expected_adapters


def test_when_data_in_adapters_and_no_inet_or_inet6_in_data_segment_then_original_data_should_be_returned():
    expected_adapters = {
        "some cool adapter": {"data": {}},
        "some other adapter": {"data": {}},
        "yet another neat adapter": {"data": {}},
    }

    no_inet_data_adapters = {"no data adapter": "this one should be gone"}
    no_inet_data_adapters.update(expected_adapters)

    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=no_inet_data_adapters
    )
    assert actual_adapters == expected_adapters


def test_when_opts_are_in_data_sorted_opt_keys_should_be_added():
    comprehensive_adapters = {
        "source": "keep me here",
        "no data adapter": "lulz",
        "adapter 1": {
            "data": {
                "inet": {
                    "ethtool": {
                        "5": {},
                        "4": {},
                        "2": {},
                        "3": {},
                        "1": {},
                    },
                    "bonding": {
                        "4": {},
                        "5": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bridging": {
                        "1": {},
                        "5": {},
                        "4": {},
                        "2": {},
                        "3": {},
                    },
                },
            },
        },
    }

    expected_adapters = {
        "source": "keep me here",
        "adapter 1": {
            "data": {
                "inet": {
                    "ethtool_keys": ["1", "2", "3", "4", "5"],
                    "bonding_keys": ["1", "2", "3", "4", "5"],
                    "bridging_keys": ["1", "2", "3", "4", "5"],
                    "ethtool": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bonding": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                    "bridging": {
                        "5": {},
                        "4": {},
                        "3": {},
                        "2": {},
                        "1": {},
                    },
                },
            },
        },
    }
    actual_adapters = debian_ip._filter_malformed_interfaces(
        adapters=comprehensive_adapters
    )
    assert actual_adapters == expected_adapters



import os
import pathlib
import sys

_repo_root = pathlib.Path(__file__).parent.parent
_paths_to_check = {""}
if sys.platform.startswith("win"):
    _paths_to_check.add(str(_repo_root).replace("\\", "\\\\"))
    _paths_to_check.add(str(_repo_root.resolve()).replace("\\", "\\\\"))
else:
    _paths_to_check.add(str(_repo_root))


if os.environ.get("ONEDIR_TESTRUN", "0") == "1":
    # In this particular case, we want to make sure that the repo root
    # is not part if sys.path so that when we import salt, we import salt from
    # the onedir and not the code checkout
    for path in _paths_to_check:
        if path in sys.path:
            sys.path.remove(path)



"""
Management of Docker networks

.. versionadded:: 2017.7.0

:depends: docker_ Python module

.. note::
    Older releases of the Python bindings for Docker were called docker-py_ in
    PyPI. All releases of docker_, and releases of docker-py_ >= 1.6.0 are
    supported. These python bindings can easily be installed using
    :py:func:`pip.install <salt.modules.pip.install>`:

    .. code-block:: bash

        salt myminion pip.install docker

    To upgrade from docker-py_ to docker_, you must first uninstall docker-py_,
    and then install docker_:

    .. code-block:: bash

        salt myminion pip.uninstall docker-py
        salt myminion pip.install docker

.. _docker: https://pypi.python.org/pypi/docker
.. _docker-py: https://pypi.python.org/pypi/docker-py

These states were moved from the :mod:`docker <salt.states.docker>` state
module (formerly called **dockerng**) in the 2017.7.0 release.
"""

import copy
import logging
import random
import string

import salt.utils.dockermod.translate.network
from salt._compat import ipaddress
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "docker_network"
__virtual_aliases__ = ("moby_network",)


def __virtual__():
    """
    Only load if the docker execution module is available
    """
    if "docker.version" in __salt__:
        return __virtualname__
    return (False, __salt__.missing_fun_string("docker.version"))


def _normalize_pools(existing, desired):
    pools = {"existing": {4: None, 6: None}, "desired": {4: None, 6: None}}

    for pool in existing["Config"]:
        subnet = ipaddress.ip_network(pool.get("Subnet"))
        pools["existing"][subnet.version] = pool

    for pool in desired["Config"]:
        subnet = ipaddress.ip_network(pool.get("Subnet"))
        if pools["desired"][subnet.version] is not None:
            raise ValueError("Only one IPv{} pool is permitted".format(subnet.version))
        else:
            pools["desired"][subnet.version] = pool

    if pools["desired"][6] and not pools["desired"][4]:
        raise ValueError(
            "An IPv4 pool is required when an IPv6 pool is used. See the "
            "documentation for details."
        )

    # The pools will be sorted when comparing
    existing["Config"] = [
        pools["existing"][x] for x in (4, 6) if pools["existing"][x] is not None
    ]
    desired["Config"] = [
        pools["desired"][x] for x in (4, 6) if pools["desired"][x] is not None
    ]


def present(
    name,
    skip_translate=None,
    ignore_collisions=False,
    validate_ip_addrs=True,
    containers=None,
    reconnect=True,
    **kwargs
):
    """
    .. versionchanged:: 2018.3.0
        Support added for network configuration options other than ``driver``
        and ``driver_opts``, as well as IPAM configuration.

    Ensure that a network is present

    .. note::
        This state supports all arguments for network and IPAM pool
        configuration which are available for the release of docker-py
        installed on the minion. For that reason, the arguments described below
        in the :ref:`NETWORK CONFIGURATION
        <salt-states-docker-network-present-netconf>` and :ref:`IP ADDRESS
        MANAGEMENT (IPAM) <salt-states-docker-network-present-ipam>` sections
        may not accurately reflect what is available on the minion. The
        :py:func:`docker.get_client_args
        <salt.modules.dockermod.get_client_args>` function can be used to check
        the available arguments for the installed version of docker-py (they
        are found in the ``network_config`` and ``ipam_config`` sections of the
        return data), but Salt will not prevent a user from attempting to use
        an argument which is unsupported in the release of Docker which is
        installed. In those cases, network creation be attempted but will fail.

    name
        Network name

    skip_translate
        This function translates Salt SLS input into the format which
        docker-py expects. However, in the event that Salt's translation logic
        fails (due to potential changes in the Docker Remote API, or to bugs in
        the translation code), this argument can be used to exert granular
        control over which arguments are translated and which are not.

        Pass this argument as a comma-separated list (or Python list) of
        arguments, and translation for each passed argument name will be
        skipped. Alternatively, pass ``True`` and *all* translation will be
        skipped.

        Skipping tranlsation allows for arguments to be formatted directly in
        the format which docker-py expects. This allows for API changes and
        other issues to be more easily worked around. See the following links
        for more information:

        - `docker-py Low-level API`_
        - `Docker Engine API`_

        .. versionadded:: 2018.3.0

    .. _`docker-py Low-level API`: http://docker-py.readthedocs.io/en/stable/api.html#docker.api.container.ContainerApiMixin.create_container
    .. _`Docker Engine API`: https://docs.docker.com/engine/api/v1.33/#operation/ContainerCreate

    ignore_collisions : False
        Since many of docker-py's arguments differ in name from their CLI
        counterparts (with which most Docker users are more familiar), Salt
        detects usage of these and aliases them to the docker-py version of
        that argument. However, if both the alias and the docker-py version of
        the same argument (e.g. ``options`` and ``driver_opts``) are used, an error
        will be raised. Set this argument to ``True`` to suppress these errors
        and keep the docker-py version of the argument.

        .. versionadded:: 2018.3.0

    validate_ip_addrs : True
        For parameters which accept IP addresses/subnets as input, validation
        will be performed. To disable, set this to ``False``.

        .. versionadded:: 2018.3.0

    containers
        A list of containers which should be connected to this network.

        .. note::
            As of the 2018.3.0 release, this is not the recommended way of
            managing a container's membership in a network, for a couple
            reasons:

            1. It does not support setting static IPs, aliases, or links in the
               container's IP configuration.
            2. If a :py:func:`docker_container.running
               <salt.states.docker_container.running>` state replaces a
               container, it will not be reconnected to the network until the
               ``docker_network.present`` state is run again. Since containers
               often have ``require`` requisites to ensure that the network
               is present, this means that the ``docker_network.present`` state
               ends up being run *before* the :py:func:`docker_container.running
               <salt.states.docker_container.running>`, leaving the container
               unattached at the end of the Salt run.

            For these reasons, it is recommended to use
            :ref:`docker_container.running's network management support
            <salt-states-docker-container-network-management>`.

    reconnect : True
        If ``containers`` is not used, and the network is replaced, then Salt
        will keep track of the containers which were connected to the network
        and reconnect them to the network after it is replaced. Salt will first
        attempt to reconnect using the same IP the container had before the
        network was replaced. If that fails (for instance, if the network was
        replaced because the subnet was modified), then the container will be
        reconnected without an explicit IP address, and its IP will be assigned
        by Docker.

        Set this option to ``False`` to keep Salt from trying to reconnect
        containers. This can be useful in some cases when :ref:`managing static
        IPs in docker_container.running
        <salt-states-docker-container-network-management>`. For instance, if a
        network's subnet is modified, it is likely that the static IP will need
        to be updated in the ``docker_container.running`` state as well. When
        the network is replaced, the initial reconnect attempt would fail, and
        the container would be reconnected with an automatically-assigned IP
        address. Then, when the ``docker_container.running`` state executes, it
        would disconnect the network *again* and reconnect using the new static
        IP. Disabling the reconnect behavior in these cases would prevent the
        unnecessary extra reconnection.

        .. versionadded:: 2018.3.0

    .. _salt-states-docker-network-present-netconf:

    **NETWORK CONFIGURATION ARGUMENTS**

    driver
        Network driver

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - driver: macvlan

    driver_opts (or *driver_opt*, or *options*)
        Options for the network driver. Either a dictionary of option names and
        values or a Python list of strings in the format ``varname=value``. The
        below three examples are equivalent:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - driver: macvlan
                - driver_opts: macvlan_mode=bridge,parent=eth0

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - driver: macvlan
                - driver_opts:
                  - macvlan_mode=bridge
                  - parent=eth0

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - driver: macvlan
                - driver_opts:
                  - macvlan_mode: bridge
                  - parent: eth0

        The options can also simply be passed as a dictionary, though this can
        be error-prone due to some :ref:`idiosyncrasies <yaml-idiosyncrasies>`
        with how PyYAML loads nested data structures:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - driver: macvlan
                - driver_opts:
                    macvlan_mode: bridge
                    parent: eth0

    check_duplicate : True
        If ``True``, checks for networks with duplicate names. Since networks
        are primarily keyed based on a random ID and not on the name, and
        network name is strictly a user-friendly alias to the network which is
        uniquely identified using ID, there is no guaranteed way to check for
        duplicates. This option providess a best effort, checking for any
        networks which have the same name, but it is not guaranteed to catch
        all name collisions.

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - check_duplicate: False

    internal : False
        If ``True``, restricts external access to the network

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - internal: True

    labels
        Add metadata to the network. Labels can be set both with and without
        values, and labels with values can be passed either as ``key=value`` or
        ``key: value`` pairs. For example, while the below would be very
        confusing to read, it is technically valid, and demonstrates the
        different ways in which labels can be passed:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - labels:
                  - foo
                  - bar=baz
                  - hello: world

        The labels can also simply be passed as a YAML dictionary, though this
        can be error-prone due to some :ref:`idiosyncrasies
        <yaml-idiosyncrasies>` with how PyYAML loads nested data structures:

        .. code-block:: yaml

            foo:
              docker_network.present:
                - labels:
                    foo: ''
                    bar: baz
                    hello: world

        .. versionchanged:: 2018.3.0
            Methods for specifying labels can now be mixed. Earlier releases
            required either labels with or without values.

    enable_ipv6 (or *ipv6*) : False
        Enable IPv6 on the network

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - enable_ipv6: True

        .. note::
            While it should go without saying, this argument must be set to
            ``True`` to :ref:`configure an IPv6 subnet
            <salt-states-docker-network-present-ipam>`. Also, if this option is
            turned on without an IPv6 subnet explicitly configured, you will
            get an error unless you have set up a fixed IPv6 subnet. Consult
            the `Docker IPv6 docs`_ for information on how to do this.

            .. _`Docker IPv6 docs`: https://docs.docker.com/v17.09/engine/userguide/networking/default_network/ipv6/

    attachable : False
        If ``True``, and the network is in the global scope, non-service
        containers on worker nodes will be able to connect to the network.

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - attachable: True

        .. note::
            This option cannot be reliably managed on CentOS 7. This is because
            while support for this option was added in API version 1.24, its
            value was not added to the inpsect results until API version 1.26.
            The version of Docker which is available for CentOS 7 runs API
            version 1.24, meaning that while Salt can pass this argument to the
            API, it has no way of knowing the value of this config option in an
            existing Docker network.

    scope
        Specify the network's scope (``local``, ``global`` or ``swarm``)

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - scope: local

    ingress : False
        If ``True``, create an ingress network which provides the routing-mesh in
        swarm mode

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ingress: True

    .. _salt-states-docker-network-present-ipam:

    **IP ADDRESS MANAGEMENT (IPAM)**

    This state supports networks with either IPv4, or both IPv4 and IPv6. If
    configuring IPv4, then you can pass the :ref:`IPAM pool arguments
    <salt-states-docker-network-present-ipam-pool-arguments>` below as
    individual arguments. However, if configuring IPv4 and IPv6, the arguments
    must be passed as a list of dictionaries, in the ``ipam_pools`` argument
    (click :ref:`here <salt-states-docker-network-present-ipam-examples>` for
    some examples). `These docs`_ also have more information on these
    arguments.

    .. _`These docs`: http://docker-py.readthedocs.io/en/stable/api.html#docker.types.IPAMPool

    *IPAM ARGUMENTS*

    ipam_driver
        IPAM driver to use, if different from the default one

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ipam_driver: foo

    ipam_opts
        Options for the IPAM driver. Either a dictionary of option names and
        values or a Python list of strings in the format ``varname=value``. The
        below three examples are equivalent:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ipam_driver: foo
                - ipam_opts: foo=bar,baz=qux

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ipam_driver: foo
                - ipam_opts:
                  - foo=bar
                  - baz=qux

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ipam_driver: foo
                - ipam_opts:
                  - foo: bar
                  - baz: qux

        The options can also simply be passed as a dictionary, though this can
        be error-prone due to some :ref:`idiosyncrasies <yaml-idiosyncrasies>`
        with how PyYAML loads nested data structures:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - ipam_driver: macvlan
                - ipam_opts:
                    foo: bar
                    baz: qux

    .. _salt-states-docker-network-present-ipam-pool-arguments:

    *IPAM POOL ARGUMENTS*

    subnet
        Subnet in CIDR format that represents a network segment

    iprange (or *ip_range*)
        Allocate container IP from a sub-range within the subnet

        Subnet in CIDR format that represents a network segment

    gateway
        IPv4 or IPv6 gateway for the master subnet

    aux_addresses (or *aux_address*)
        A dictionary of mapping container names to IP addresses which should be
        allocated for them should they connect to the network. Either a
        dictionary of option names and values or a Python list of strings in
        the format ``host=ipaddr``.

    .. _salt-states-docker-network-present-ipam-examples:

    *IPAM CONFIGURATION EXAMPLES*

    Below is an example of an IPv4-only network (keep in mind that ``subnet``
    is the only required argument).

    .. code-block:: yaml

        mynet:
          docker_network.present:
            - subnet: 10.0.20.0/24
            - iprange: 10.0.20.128/25
            - gateway: 10.0.20.254
            - aux_addresses:
              - foo.bar.tld: 10.0.20.50
              - hello.world.tld: 10.0.20.51

    .. note::
        The ``aux_addresses`` can be passed differently, in the same way that
        ``driver_opts`` and ``ipam_opts`` can.

    This same network could also be configured this way:

    .. code-block:: yaml

        mynet:
          docker_network.present:
            - ipam_pools:
              - subnet: 10.0.20.0/24
                iprange: 10.0.20.128/25
                gateway: 10.0.20.254
                aux_addresses:
                  foo.bar.tld: 10.0.20.50
                  hello.world.tld: 10.0.20.51

    Here is an example of a mixed IPv4/IPv6 subnet.

    .. code-block:: yaml

        mynet:
          docker_network.present:
            - ipam_pools:
              - subnet: 10.0.20.0/24
                gateway: 10.0.20.1
              - subnet: fe3f:2180:26:1::/123
                gateway: fe3f:2180:26:1::1
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    try:
        network = __salt__["docker.inspect_network"](name)
    except CommandExecutionError as exc:
        msg = exc.__str__()
        if "404" in msg:
            # Network not present
            network = None
        else:
            ret["comment"] = msg
            return ret

    # map container's IDs to names
    to_connect = {}
    missing_containers = []
    stopped_containers = []
    for cname in __utils__["args.split_input"](containers or []):
        try:
            cinfo = __salt__["docker.inspect_container"](cname)
        except CommandExecutionError:
            missing_containers.append(cname)
        else:
            try:
                cid = cinfo["Id"]
            except KeyError:
                missing_containers.append(cname)
            else:
                if not cinfo.get("State", {}).get("Running", False):
                    stopped_containers.append(cname)
                else:
                    to_connect[cid] = {"Name": cname}

    if missing_containers:
        ret.setdefault("warnings", []).append(
            "The following containers do not exist: {}.".format(
                ", ".join(missing_containers)
            )
        )

    if stopped_containers:
        ret.setdefault("warnings", []).append(
            "The following containers are not running: {}.".format(
                ", ".join(stopped_containers)
            )
        )

    # We might disconnect containers in the process of recreating the network,
    # we'll need to keep track these containers so we can reconnect them later.
    disconnected_containers = {}

    try:
        kwargs = __utils__["docker.translate_input"](
            salt.utils.dockermod.translate.network,
            skip_translate=skip_translate,
            ignore_collisions=ignore_collisions,
            validate_ip_addrs=validate_ip_addrs,
            **__utils__["args.clean_kwargs"](**kwargs)
        )
    except Exception as exc:  # pylint: disable=broad-except
        ret["comment"] = exc.__str__()
        return ret

    # Separate out the IPAM config options and build the IPAM config dict
    ipam_kwargs = {}
    ipam_kwarg_names = ["ipam", "ipam_driver", "ipam_opts", "ipam_pools"]
    ipam_kwarg_names.extend(
        __salt__["docker.get_client_args"]("ipam_config")["ipam_config"]
    )
    for key in ipam_kwarg_names:
        try:
            ipam_kwargs[key] = kwargs.pop(key)
        except KeyError:
            pass
    if "ipam" in ipam_kwargs:
        if len(ipam_kwargs) > 1:
            ret["comment"] = (
                "Cannot mix the 'ipam' argument with any of the IPAM config "
                "arguments. See documentation for details."
            )
            return ret
        ipam_config = ipam_kwargs["ipam"]
    else:
        ipam_pools = ipam_kwargs.pop("ipam_pools", ())
        try:
            ipam_config = __utils__["docker.create_ipam_config"](
                *ipam_pools, **ipam_kwargs
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = exc.__str__()
            return ret

    # We'll turn this off if we decide below that creating the network is not
    # necessary.
    create_network = True

    if network is not None:
        log.debug("Docker network '%s' already exists", name)

        # Set the comment now to say that it already exists, if we need to
        # recreate the network with new config we'll update the comment later.
        ret[
            "comment"
        ] = "Network '{}' already exists, and is configured as specified".format(name)
        log.trace("Details of docker network '%s': %s", name, network)

        temp_net_name = "".join(
            random.choice(string.ascii_lowercase) for _ in range(20)
        )

        try:
            # When using enable_ipv6, you *must* provide a subnet. But we don't
            # care about the subnet when we make our temp network, we only care
            # about the non-IPAM values in the network. And we also do not want
            # to try some hacky workaround where we choose a small IPv6 subnet
            # to pass when creating the temp network, that may end up
            # overlapping with a large IPv6 subnet already in use by Docker.
            # So, for purposes of comparison we will create the temp network
            # with enable_ipv6=False and then munge the inspect results before
            # performing the comparison. Note that technically it is not
            # required that one specify both v4 and v6 subnets when creating a
            # network, but not specifying IPv4 makes it impossible for us to
            # reliably compare the SLS input to the existing network, as we
            # wouldng't know if the IPv4 subnet in the existing network was
            # explicitly configured or was automatically assigned by Docker.
            enable_ipv6 = kwargs.pop("enable_ipv6", None)
            kwargs_tmp = kwargs
            driver = kwargs.get(
                "driver",
            )
            driver_opts = kwargs.get("options", {})
            bridge_name = driver_opts.get("com.docker.network.bridge.name", None)

            if driver == "bridge" and bridge_name is not None:
                tmp_name = str(bridge_name) + "comp"
                kwargs_tmp["options"]["com.docker.network.bridge.name"] = tmp_name[-14:]
            __salt__["docker.create_network"](
                temp_net_name,
                skip_translate=True,  # No need to translate (already did)
                enable_ipv6=False,
                **kwargs_tmp
            )
        except CommandExecutionError as exc:
            ret["comment"] = "Failed to create temp network for comparison: {}".format(
                exc.__str__()
            )
            return ret
        else:
            # Replace the value so we can use it later
            if enable_ipv6 is not None:
                kwargs["enable_ipv6"] = enable_ipv6

        try:
            try:
                temp_net_info = __salt__["docker.inspect_network"](temp_net_name)
            except CommandExecutionError as exc:
                ret["comment"] = "Failed to inspect temp network: {}".format(
                    exc.__str__()
                )
                return ret
            else:
                temp_net_info["EnableIPv6"] = bool(enable_ipv6)

            # Replace the IPAM configuration in the temp network with the IPAM
            # config dict we created earlier, for comparison purposes. This is
            # necessary because we cannot create two networks that have
            # overlapping subnets (the Docker Engine will throw an error).
            temp_net_info["IPAM"] = ipam_config

            existing_pool_count = len(network["IPAM"]["Config"])
            desired_pool_count = len(temp_net_info["IPAM"]["Config"])

            is_default_pool = (
                lambda x: True if sorted(x) == ["Gateway", "Subnet"] else False
            )

            if (
                desired_pool_count == 0
                and existing_pool_count == 1
                and is_default_pool(network["IPAM"]["Config"][0])
            ):
                # If we're not explicitly configuring an IPAM pool, then we
                # don't care what the subnet is. Docker networks created with
                # no explicit IPAM configuration are assigned a single IPAM
                # pool containing just a subnet and gateway. If the above if
                # statement resolves as True, then we know that both A) we
                # aren't explicitly configuring IPAM, and B) the existing
                # network appears to be one that was created without an
                # explicit IPAM configuration (since it has the default pool
                # config values). Of course, it could be possible that the
                # existing network was created with a single custom IPAM pool,
                # with just a subnet and gateway. But even if this was the
                # case, the fact that we aren't explicitly enforcing IPAM
                # configuration means we don't really care what the existing
                # IPAM configuration is. At any rate, to avoid IPAM differences
                # when comparing the existing network to the temp network, we
                # need to clear the existing network's IPAM configuration.
                network["IPAM"]["Config"] = []

            changes = __salt__["docker.compare_networks"](
                network, temp_net_info, ignore="Name,Id,Created,Containers"
            )

            if not changes:
                # No changes to the network, so we'll be keeping the existing
                # network and at most just connecting containers to it.
                create_network = False

            else:
                ret["changes"][name] = changes
                if __opts__["test"]:
                    ret["result"] = None
                    ret["comment"] = "Network would be recreated with new config"
                    return ret

                if network["Containers"]:
                    # We've removed the network, so there are now no containers
                    # attached to it. However, once we recreate the network
                    # with the new configuration we may need to reconnect the
                    # containers that were previously connected. Even if we're
                    # not reconnecting, we still need to track the containers
                    # so that we can report on which were disconnected.
                    disconnected_containers = copy.deepcopy(network["Containers"])
                    if not containers and reconnect:
                        # Grab the links and aliases from each connected
                        # container so that we have them when we attempt to
                        # reconnect later
                        for cid in disconnected_containers:
                            try:
                                cinfo = __salt__["docker.inspect_container"](cid)
                                netinfo = cinfo["NetworkSettings"]["Networks"][name]
                                # Links and Aliases will be None if not
                                # explicitly set, hence using "or" instead of
                                # placing the empty list inside the dict.get
                                net_links = netinfo.get("Links") or []
                                net_aliases = netinfo.get("Aliases") or []
                                if net_links:
                                    disconnected_containers[cid]["Links"] = net_links
                                if net_aliases:
                                    disconnected_containers[cid][
                                        "Aliases"
                                    ] = net_aliases
                            except (CommandExecutionError, KeyError, ValueError):
                                continue

                remove_result = _remove_network(network)
                if not remove_result["result"]:
                    return remove_result

                # Replace the Containers key with an empty dict so that when we
                # check for connnected containers below, we correctly see that
                # there are none connected.
                network["Containers"] = {}
        finally:
            try:
                __salt__["docker.remove_network"](temp_net_name)
            except CommandExecutionError as exc:
                ret.setdefault("warnings", []).append(
                    "Failed to remove temp network '{}': {}.".format(
                        temp_net_name, exc.__str__()
                    )
                )

    if create_network:
        log.debug("Network '%s' will be created", name)
        if __opts__["test"]:
            # NOTE: if the container already existed and needed to be
            # recreated, and we were in test mode, we would have already exited
            # above with a comment about the network needing to be recreated.
            # So, even though the below block to create the network would be
            # executed to create the network both when it's being recreated and
            # when it's being created for the first time, the below comment is
            # still accurate.
            ret["result"] = None
            ret["comment"] = "Network will be created"
            return ret

        kwargs["ipam"] = ipam_config
        try:
            __salt__["docker.create_network"](
                name,
                skip_translate=True,  # No need to translate (already did)
                **kwargs
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Failed to create network '{}': {}".format(
                name, exc.__str__()
            )
            return ret
        else:
            action = "recreated" if network is not None else "created"
            ret["changes"][action] = True
            ret["comment"] = "Network '{}' {}".format(
                name,
                "created" if network is None else "was replaced with updated config",
            )
            # Make sure the "Containers" key exists for logic below
            network = {"Containers": {}}

    # If no containers were specified in the state but we have disconnected
    # some in the process of recreating the network, we should reconnect those
    # containers.
    if containers is None and reconnect and disconnected_containers:
        to_connect = disconnected_containers

    # Don't try to connect any containers which are already connected. If we
    # created/re-created the network, then network['Containers'] will be empty
    # and no containers will be deleted from the to_connect dict (the result
    # being that we will reconnect all containers in the to_connect dict).
    # list() is used here because we will potentially be modifying the
    # dictionary during iteration.
    for cid in list(to_connect):
        if cid in network["Containers"]:
            del to_connect[cid]

    errors = []
    if to_connect:
        for cid, connect_info in to_connect.items():
            connect_kwargs = {}
            if cid in disconnected_containers:
                for key_name, arg_name in (
                    ("IPv4Address", "ipv4_address"),
                    ("IPV6Address", "ipv6_address"),
                    ("Links", "links"),
                    ("Aliases", "aliases"),
                ):
                    try:
                        connect_kwargs[arg_name] = connect_info[key_name]
                    except (KeyError, AttributeError):
                        continue
                    else:
                        if key_name.endswith("Address"):
                            connect_kwargs[arg_name] = connect_kwargs[arg_name].rsplit(
                                "/", 1
                            )[0]
            try:
                __salt__["docker.connect_container_to_network"](
                    cid, name, **connect_kwargs
                )
            except CommandExecutionError as exc:
                if not connect_kwargs:
                    errors.append(exc.__str__())
                else:
                    # We failed to reconnect with the container's old IP
                    # configuration. Reconnect using automatic IP config.
                    try:
                        __salt__["docker.connect_container_to_network"](cid, name)
                    except CommandExecutionError as exc:
                        errors.append(exc.__str__())
                    else:
                        ret["changes"].setdefault(
                            "reconnected"
                            if cid in disconnected_containers
                            else "connected",
                            [],
                        ).append(connect_info["Name"])
            else:
                ret["changes"].setdefault(
                    "reconnected" if cid in disconnected_containers else "connected", []
                ).append(connect_info["Name"])

    if errors:
        if ret["comment"]:
            ret["comment"] += ". "
        ret["comment"] += ". ".join(errors) + "."
    else:
        ret["result"] = True

    # Figure out if we removed any containers as a result of replacing the
    # network and did not reconnect them. We only would not have reconnected if
    # a list of containers was passed in the "containers" argument, and there
    # were containers connected to the network prior to its replacement which
    # were not part of that list.
    for cid, c_info in disconnected_containers.items():
        if cid not in to_connect:
            ret["changes"].setdefault("disconnected", []).append(c_info["Name"])

    return ret


def absent(name):
    """
    Ensure that a network is absent.

    name
        Name of the network

    Usage Example:

    .. code-block:: yaml

        network_foo:
          docker_network.absent
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    try:
        network = __salt__["docker.inspect_network"](name)
    except CommandExecutionError as exc:
        msg = exc.__str__()
        if "404" in msg:
            # Network not present
            network = None
        else:
            ret["comment"] = msg
            return ret

    if network is None:
        ret["result"] = True
        ret["comment"] = "Network '{}' already absent".format(name)
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Network '{}' will be removed".format(name)
        return ret

    return _remove_network(network)


def _remove_network(network):
    """
    Remove network, including all connected containers
    """
    ret = {"name": network["Name"], "changes": {}, "result": False, "comment": ""}

    errors = []
    for cid in network["Containers"]:
        try:
            cinfo = __salt__["docker.inspect_container"](cid)
        except CommandExecutionError:
            # Fall back to container ID
            cname = cid
        else:
            cname = cinfo.get("Name", "").lstrip("/")

        try:
            __salt__["docker.disconnect_container_from_network"](cid, network["Name"])
        except CommandExecutionError as exc:
            errors = "Failed to disconnect container '{}' : {}".format(cname, exc)
        else:
            ret["changes"].setdefault("disconnected", []).append(cname)

    if errors:
        ret["comment"] = "\n".join(errors)
        return ret

    try:
        __salt__["docker.remove_network"](network["Name"])
    except CommandExecutionError as exc:
        ret["comment"] = "Failed to remove network: {}".format(exc)
    else:
        ret["changes"]["removed"] = True
        ret["result"] = True
        ret["comment"] = "Removed network '{}'".format(network["Name"])

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
    :codeauthor: Gareth J. Greenaway <ggreenaway@vmware.com>
"""

import pytest

import salt.modules.schedule as schedule_mod
import salt.states.schedule as schedule
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {schedule: {}}


def test_present():
    """
    Test to ensure a job is present in the schedule.
    """
    name = "job1"

    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    mock_lst = MagicMock(side_effect=[{}, {"job1": job1}])

    mock_build_schedule = OrderedDict(
        [
            ("function", "test.ping"),
            ("maxrunning", 1),
            ("name", "job1"),
            ("enabled", True),
            ("jid_include", True),
            ("when", "4:00am"),
        ]
    )

    mock_add = {
        "comment": "Added job: test-schedule to schedule.",
        "result": True,
        "changes": {"test-schedule": "added"},
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.build_schedule_item": MagicMock(return_value=mock_build_schedule),
            "schedule.add": MagicMock(return_value=mock_add),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {"test-schedule": "added"},
            "comment": "Adding new job job1 to schedule",
        }
        _res = schedule.present(name)
        assert _res == ret

        ret = {
            "name": "job1",
            "result": True,
            "changes": {},
            "comment": "Job job1 in correct state",
        }
        _res = schedule.present(name)
        assert _res == ret

    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    job1_update = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "6:00am",
    }
    mock_lst = MagicMock(side_effect=[{"job1": job1}, {"job1": job1_update}])

    mock_build_schedule = OrderedDict(
        [
            ("function", "test.ping"),
            ("maxrunning", 1),
            ("name", "job1"),
            ("enabled", True),
            ("jid_include", True),
            ("when", "6:00am"),
        ]
    )

    mock_modify = {
        "comment": "Modified job: test-schedule in schedule.",
        "changes": {
            "test-schedule": {
                "old": OrderedDict(
                    [
                        ("function", "test.ping"),
                        ("maxrunning", 1),
                        ("name", "test-schedule"),
                        ("enabled", True),
                        ("jid_include", True),
                        ("when", "4:00am"),
                    ]
                ),
                "new": OrderedDict(
                    [
                        ("function", "test.ping"),
                        ("maxrunning", 1),
                        ("name", "test-schedule"),
                        ("enabled", True),
                        ("jid_include", True),
                        ("when", "6:00am"),
                    ]
                ),
            }
        },
        "result": True,
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.build_schedule_item": MagicMock(return_value=mock_build_schedule),
            "schedule.modify": MagicMock(return_value=mock_modify),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {
                "test-schedule": {
                    "old": OrderedDict(
                        [
                            ("function", "test.ping"),
                            ("maxrunning", 1),
                            ("name", "test-schedule"),
                            ("enabled", True),
                            ("jid_include", True),
                            ("when", "4:00am"),
                        ]
                    ),
                    "new": OrderedDict(
                        [
                            ("function", "test.ping"),
                            ("maxrunning", 1),
                            ("name", "test-schedule"),
                            ("enabled", True),
                            ("jid_include", True),
                            ("when", "6:00am"),
                        ]
                    ),
                }
            },
            "comment": "Modifying job job1 in schedule",
        }
        _res = schedule.present(name)
        assert _res == ret

        ret = {
            "name": "job1",
            "result": True,
            "changes": {},
            "comment": "Job job1 in correct state",
        }
        _res = schedule.present(name)
        assert _res == ret

    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    mock_lst = MagicMock(side_effect=[{}])

    mock_build_schedule = OrderedDict(
        [
            ("function", "test.ping"),
            ("maxrunning", 1),
            ("name", "job1"),
            ("enabled", True),
            ("jid_include", True),
            ("when", "4:00am"),
        ]
    )

    mock_add = {
        "comment": "Job: test-schedule would be added to schedule.",
        "result": True,
        "changes": {},
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.build_schedule_item": MagicMock(return_value=mock_build_schedule),
            "schedule.add": MagicMock(return_value=mock_add),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {},
            "comment": "Job: test-schedule would be added to schedule.",
        }
        with patch.dict(schedule.__opts__, {"test": True}):
            _res = schedule.present(name)
            assert _res == ret

    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    job1_update = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "6:00am",
    }
    mock_lst = MagicMock(side_effect=[{"job1": job1}, {"job1": job1_update}])

    mock_build_schedule = OrderedDict(
        [
            ("function", "test.ping"),
            ("maxrunning", 1),
            ("name", "job1"),
            ("enabled", True),
            ("jid_include", True),
            ("when", "6:00am"),
        ]
    )

    mock_modify = {
        "comment": "Job: test-schedule would be modified in schedule.",
        "changes": {
            "test-schedule": {
                "old": OrderedDict(
                    [
                        ("function", "test.ping"),
                        ("maxrunning", 1),
                        ("name", "test-schedule"),
                        ("enabled", True),
                        ("jid_include", True),
                        ("when", "4:00am"),
                    ]
                ),
                "new": OrderedDict(
                    [
                        ("function", "test.ping"),
                        ("maxrunning", 1),
                        ("name", "test-schedule"),
                        ("enabled", True),
                        ("jid_include", True),
                        ("when", "6:00am"),
                    ]
                ),
            }
        },
        "result": True,
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.build_schedule_item": MagicMock(return_value=mock_build_schedule),
            "schedule.modify": MagicMock(return_value=mock_modify),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {
                "test-schedule": {
                    "old": OrderedDict(
                        [
                            ("function", "test.ping"),
                            ("maxrunning", 1),
                            ("name", "test-schedule"),
                            ("enabled", True),
                            ("jid_include", True),
                            ("when", "4:00am"),
                        ]
                    ),
                    "new": OrderedDict(
                        [
                            ("function", "test.ping"),
                            ("maxrunning", 1),
                            ("name", "test-schedule"),
                            ("enabled", True),
                            ("jid_include", True),
                            ("when", "6:00am"),
                        ]
                    ),
                }
            },
            "comment": "Job: test-schedule would be modified in schedule.",
        }
        with patch.dict(schedule.__opts__, {"test": True}):
            _res = schedule.present(name)
            assert _res == ret

    # Add job to schedule when offline=True
    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
        "offline": True,
    }
    mock_lst = MagicMock(return_value={})

    mock_build_schedule = OrderedDict(
        [
            ("function", "test.ping"),
            ("maxrunning", 1),
            ("name", "job1"),
            ("enabled", True),
            ("jid_include", True),
            ("when", "4:00am"),
        ]
    )

    mock_add = {
        "comment": "Adding new job test-schedule to schedule.",
        "result": True,
        "changes": {"test-schedule": "added"},
    }

    event_enter = MagicMock()
    event_enter.send.side_effect = (lambda data, tag, cb=None, timeout=60: True,)
    event = MagicMock()
    event.__enter__.return_value = event_enter

    with patch("salt.utils.event.get_event", return_value=event):
        with patch.dict(
            schedule.__salt__,
            {
                "schedule.list": mock_lst,
                "schedule.build_schedule_item": MagicMock(
                    return_value=mock_build_schedule
                ),
                "schedule.add": MagicMock(return_value=mock_add),
            },
        ):
            with patch.object(schedule_mod, "list_", mock_lst):
                with patch.object(
                    schedule_mod,
                    "_get_schedule_config_file",
                    MagicMock(return_value="/etc/salt/minion.d/_schedule.conf"),
                ):
                    with patch("salt.utils.files.fopen", mock_open()):
                        ret = {
                            "comment": "Adding new job job1 to schedule",
                            "result": True,
                            "name": "job1",
                            "changes": {"test-schedule": "added"},
                        }

                        _res = schedule.present(name, offline=True)
                        assert _res == ret
                        assert event.call_count == 0


def test_absent():
    """
    Test to ensure a job is absent from the schedule.
    """

    # Delete job from schedule
    name = "job1"

    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    mock_lst = MagicMock(side_effect=[{"job1": job1}])

    mock_delete = {
        "comment": "Deleted job test-schedule from schedule.",
        "result": True,
        "changes": {"test-schedule": "removed"},
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.delete": MagicMock(return_value=mock_delete),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {"test-schedule": "removed"},
            "comment": "Removed job job1 from schedule",
        }
        _res = schedule.absent(name)
        assert _res == ret

    # Delete job from schedule when job does not exist
    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    mock_lst = MagicMock(side_effect=[{}])

    mock_delete = {
        "comment": "Job test-schedule does not exist.",
        "result": True,
        "changes": {},
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.delete": MagicMock(return_value=mock_delete),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {},
            "comment": "Job job1 not present in schedule",
        }
        _res = schedule.absent(name)
        assert _res == ret

    # Delete job from schedule when test=True
    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
    }
    mock_lst = MagicMock(side_effect=[{"job1": job1}])

    mock_delete = {
        "comment": "Job: job1 would be deleted from schedule.",
        "result": True,
        "changes": {},
    }

    with patch.dict(
        schedule.__salt__,
        {
            "schedule.list": mock_lst,
            "schedule.delete": MagicMock(return_value=mock_delete),
        },
    ):
        ret = {
            "name": "job1",
            "result": True,
            "changes": {},
            "comment": "Job: job1 would be deleted from schedule.",
        }

        with patch.dict(schedule.__opts__, {"test": True}):
            _res = schedule.absent(name)
            assert _res == ret

    # Delete job from schedule when offline=True
    job1 = {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
        "when": "4:00am",
        "offline": True,
    }
    mock_lst = MagicMock(return_value={"job1": job1})

    mock_delete = {
        "comment": "Deleted Job job1 from schedule.",
        "result": True,
        "changes": {"job1": "removed"},
    }

    event_enter = MagicMock()
    event_enter.send.side_effect = (lambda data, tag, cb=None, timeout=60: True,)
    event = MagicMock()
    event.__enter__.return_value = event_enter

    with patch("salt.utils.event.get_event", return_value=event):
        with patch.dict(
            schedule.__salt__,
            {
                "schedule.list": mock_lst,
                "schedule.delete": schedule_mod.delete,
            },
        ):
            with patch.object(schedule_mod, "list_", mock_lst):
                with patch.object(
                    schedule_mod,
                    "_get_schedule_config_file",
                    MagicMock(return_value="/etc/salt/minion.d/_schedule.conf"),
                ):
                    with patch("salt.utils.files.fopen", mock_open()):
                        ret = {
                            "comment": "Removed job job1 from schedule",
                            "result": True,
                            "name": "job1",
                            "changes": {"job1": "removed"},
                        }

                        _res = schedule.absent(name, offline=True)
                        assert _res == ret
                        assert event.call_count == 0



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_api(api_request):
    """
    Test running a command against the salt api
    """
    ret = api_request.post(
        "/run",
        data={
            "client": "local",
            "tgt": "*",
            "fun": "test.arg",
            "arg": ["foo", "bar"],
        },
    )
    assert ret["args"] == ["foo", "bar"]
